#include <rt_config.h>

// Radar detection and carrier detection for 2880-SW-MCU

#if defined(DFS_SUPPORT) || defined(CARRIER_DETECTION_SUPPORT)

static inline void TimerCB_Radar(PRTMP_ADAPTER pAd);
static inline void TimerCB_Carrier(PRTMP_ADAPTER pAd);

void MCURadarDetect(PRTMP_ADAPTER pAd);

void TimerCB(PRTMP_ADAPTER pAd)
{
	// what is this mean???
	if (pAd->CommonCfg.McuRadarProtection == 1)
	{
#ifdef DFS_DEBUG
		printk("###\n");
#endif // DFS_DEBUG //
		return;
	}

#ifdef NEW_DFS
	if ((pAd->MACVersion == 0x28720200) && (pAd->CommonCfg.CID == 0x200))
	{
		NewTimerCB_Radar(pAd);
	}
	else
#endif // NEW_DFS //


#ifdef CONFIG_AP_SUPPORT	
#ifdef DFS_SUPPORT
	if (pAd->CommonCfg.McuRadarCmd & RADAR_DETECTION)
		TimerCB_Radar(pAd);
#endif
#endif // CONFIG_AP_SUPPORT //

#ifdef CARRIER_DETECTION_SUPPORT
	if (pAd->CommonCfg.McuRadarCmd & CARRIER_DETECTION)
		TimerCB_Carrier(pAd);
#endif

}

#ifdef DFS_DEBUG
static unsigned int dbg_idx=0;

#define DFS_DEBUG_BEGIN				0x1
#define DFS_DEBUG_CTS_KICK			0x2
#define DFS_DEBUG_CTS_SENT			0x3
#define DFS_DEBUG_WAIT_CTS			0x4
#define DFS_DEBUG_DETECT_NO_CTS		0x5
#define DFS_DEBUG_DETECT_AFTER_CTS	0x6
#define DFS_DEBUG_WAIT_CTS_TIMEOUT	0x7
#define DFS_DEBUG_IN_DETECTION		0x8
#define DFS_DEBUG_FREE_TX			0x9
#define DFS_DEBUG_LONG_REPORT		0x10
#define DFS_DEBUG_SHORT_REPORT		0x11
#define DFS_DEBUG_SWITCH_PRI		0x12
#define DFS_DEBUG_SWITCH_EXT		0x13
#define DFS_DEBUG_0001				0x14


void DFS_Debug(UCHAR i, PRTMP_ADAPTER pAd)
{
	if (dbg_idx < 100)
	{
		pAd->CommonCfg.Detect1msIndex[dbg_idx] = i;
		pAd->CommonCfg.Detect1msTick[dbg_idx] = pAd->CommonCfg.McuRadarTick;
		pAd->CommonCfg.Detect1msDetectCount[dbg_idx] = pAd->CommonCfg.McuRadarDetectCount;
		pAd->CommonCfg.Detect1msState[dbg_idx] = pAd->CommonCfg.McuRadarState;
		pAd->CommonCfg.Detect1msEvent[dbg_idx] = pAd->CommonCfg.McuRadarEvent;

		RTMP_IO_READ32(pAd, PBF_LIFE_TIMER, &pAd->CommonCfg.Detect1msTime[dbg_idx]);

		dbg_idx++;
	}
	if (dbg_idx == 100)
	{
		printk("data collect done\n");
		dbg_idx++;
	}
}

void DFS_Dump(PRTMP_ADAPTER pAd)
{
	int i;
	for (i = 0; i < 100; i++)
	{
		printk("%02x : %02d %d ", pAd->CommonCfg.Detect1msIndex[i], pAd->CommonCfg.Detect1msTick[i], pAd->CommonCfg.Detect1msDetectCount[i]);
		printk("%d %x (%x)\n", pAd->CommonCfg.Detect1msState[i], pAd->CommonCfg.Detect1msEvent[i], (pAd->CommonCfg.Detect1msTime[i] - ((i == 0)? 0 : pAd->CommonCfg.Detect1msTime[i-1])));
	}
	
	dbg_idx=0;
}
#endif // DFS_DEBUG //

#define CTS_WAIT_LOOP		500

#ifdef CONFIG_AP_SUPPORT
#ifdef DFS_SUPPORT
static inline void TimerCB_Radar(PRTMP_ADAPTER pAd)
{
	ULONG Value;

	if (pAd->CommonCfg.McuRadarEvent & RADAR_EVENT_CARRIER_DETECTING)
	{
		pAd->CommonCfg.McuRadarTick++;
		return;
	}
	else
		pAd->CommonCfg.McuRadarTick++;
	

	if ((pAd->CommonCfg.McuRadarTick >= pAd->CommonCfg.McuRadarPeriod) && (pAd->CommonCfg.McuRadarState != DO_DETECTION))
	{

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(DFS_DEBUG_BEGIN, pAd);
			}
#endif // DFS_DEBUG //

		/* Roger add to fix false detection(long pulse only) in the first 60 seconds */
		if (pAd->CommonCfg.W56_debug)
		{
			ULONG time;
			RTMP_IO_READ32(pAd, PBF_LIFE_TIMER, &pAd->CommonCfg.W56_hw_2);
			RTMP_IO_READ32(pAd, CH_IDLE_STA, &Value);
			
			pAd->CommonCfg.W56_hw_sum += pAd->CommonCfg.W56_hw_2 - pAd->CommonCfg.W56_hw_1;
			pAd->CommonCfg.W56_hw_1 = pAd->CommonCfg.W56_hw_2;
			pAd->CommonCfg.W56_idx++;
			pAd->CommonCfg.W56_1s += Value;
			if ((pAd->CommonCfg.W56_idx % 100) == 0)
			{
				int i;
				//printk("~~~ %d   %d   ", pAd->CommonCfg.W56_idx, pAd->CommonCfg.W56_hw_sum - pAd->CommonCfg.W56_1s);
				if (pAd->CommonCfg.W56_hw_sum > pAd->CommonCfg.W56_1s)
					pAd->CommonCfg.W56_4s[(pAd->CommonCfg.W56_idx / 100) % 4] = pAd->CommonCfg.W56_hw_sum - pAd->CommonCfg.W56_1s;
				else
					pAd->CommonCfg.W56_4s[(pAd->CommonCfg.W56_idx / 100) % 4] = 0;
				Value = 0;
				for (i = 0; i < 4; i++)
					Value += pAd->CommonCfg.W56_4s[i];
				//printk("%d\n", Value);
				pAd->CommonCfg.W56_total = Value;
				pAd->CommonCfg.W56_hw_sum = 0;
				pAd->CommonCfg.W56_1s = 0;
			}
		}

		// Start of period
		pAd->CommonCfg.McuRadarTick = 0;
		pAd->CommonCfg.McuRadarDetectCount = 0;
		pAd->CommonCfg.McuRadarState = WAIT_CTS_BEING_SENT;
		pAd->CommonCfg.McuRadarEvent |= RADAR_EVENT_RADAR_DETECTING;


		if ((pAd->CommonCfg.McuRadarCtsProtect == 1)
#ifdef CARRIER_DETECTION_SUPPORT
			|| (isCarrierDetectExist(pAd))
#endif
			)
		{
			// no CTS protect
			pAd->CommonCfg.McuRadarState = DO_DETECTION;
			MCURadarDetect(pAd);

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(DFS_DEBUG_DETECT_NO_CTS, pAd);
			}
#endif // DFS_DEBUG //
			
		}
		else if (pAd->CommonCfg.McuRadarCtsProtect == 2 || pAd->CommonCfg.McuRadarCtsProtect == 3)
		{
			int i;
						
			if (pAd->CommonCfg.McuRadarCtsProtect == 3)
			{
				
				RTMP_IO_READ32(pAd, 0x7784, &Value);
				Value &= 0xffff00ff;
				RTMP_IO_WRITE32(pAd, 0x7784, Value);
				
				// pAd->CommonCfg.McuRadarCtsProtect == 3, need to kick CTS two times
				RTMP_IO_WRITE32(pAd, PBF_CTRL, 0x40);
				for (i = 0; i < CTS_WAIT_LOOP; i++)
				{
					RTMP_IO_READ32(pAd, PBF_CTRL, &Value);
					if (!Value)
						break;
				}
				
				if (i >= CTS_WAIT_LOOP)
				{
					pAd->CommonCfg.McuRadarEvent |= RADAR_EVENT_CTS_KICKED;
					return;
				}
			}


			RTMP_IO_READ32(pAd, 0x7784, &Value);
			Value &= 0xffff00ff;
			Value |= (DFS_CTS_WCID << 8);
			RTMP_IO_WRITE32(pAd, 0x7784, Value);

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(DFS_DEBUG_CTS_KICK, pAd);
			}
#endif // DFS_DEBUG //
			// kick CTS
			RTMP_IO_WRITE32(pAd, PBF_CTRL, 0x40);

		}
		return;
	}

	if (pAd->CommonCfg.McuRadarState == WAIT_CTS_BEING_SENT)
	{

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(DFS_DEBUG_WAIT_CTS, pAd);
			}
#endif // DFS_DEBUG //

		// check event or timeout
		if (pAd->CommonCfg.McuRadarEvent & RADAR_EVENT_CTS_SENT)
		{
			pAd->CommonCfg.McuRadarEvent &= ~RADAR_EVENT_CTS_SENT;
			pAd->CommonCfg.McuRadarDetectCount = 1;
			pAd->CommonCfg.McuRadarState = DO_DETECTION;

			{
				UCHAR BBPR115;
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R115, &BBPR115);
			}

#ifdef DFS_DEBUG			
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(DFS_DEBUG_DETECT_AFTER_CTS, pAd);
			}
#endif // DFS_DEBUG //
			return;
		}
		
		if (pAd->CommonCfg.McuRadarDetectCount >= (4-1)) // timeout 4ms, start from 0
		{
			pAd->CommonCfg.McuRadarState = FREE_FOR_TX; // timeout 5ms, give up
			pAd->CommonCfg.McuRadarEvent &= ~(RADAR_EVENT_RADAR_DETECTING | RADAR_EVENT_CTS_KICKED);
			pAd->CommonCfg.McuRadarTick = pAd->CommonCfg.McuRadarPeriod;

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(DFS_DEBUG_WAIT_CTS_TIMEOUT, pAd);
			}
#endif // DFS_DEBUG //

			RTMP_IO_WRITE32(pAd, TX_RTS_CFG, pAd->CommonCfg.OldRtsRetryLimit);
			// Restore BBP
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R65, pAd->CommonCfg._R65);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, pAd->CommonCfg._R66);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x01);

			// EnableNormalTx
			RTMP_IO_READ32(pAd, MAC_SYS_CTRL, &Value);
			Value |= 0x04;
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, Value);
			
		}

		if (pAd->CommonCfg.McuRadarEvent & RADAR_EVENT_CTS_KICKED)
		{
			int i;

			for (i = 0; i < CTS_WAIT_LOOP; i++)
			{
				RTMP_IO_READ32(pAd, PBF_CTRL, &Value);
				if (!Value)
					break;
			}
				
			if (i >= CTS_WAIT_LOOP)
			{
				pAd->CommonCfg.McuRadarDetectCount++;
				return;
			}
			
			RTMP_IO_READ32(pAd, 0x7784, &Value);
			Value &= 0xffff00ff;
			Value |= (DFS_CTS_WCID << 8) ;
			RTMP_IO_WRITE32(pAd, 0x7784, Value);

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(DFS_DEBUG_CTS_KICK, pAd);
			}
#endif // DFS_DEBUG //

			// kick CTS
			RTMP_IO_WRITE32(pAd, PBF_CTRL, 0x40);
			
			pAd->CommonCfg.McuRadarEvent &= ~RADAR_EVENT_CTS_KICKED;
		}

		
		pAd->CommonCfg.McuRadarDetectCount++;
	}
	else if (pAd->CommonCfg.McuRadarState == DO_DETECTION)
	{
		UCHAR BBPR115;
		

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(DFS_DEBUG_IN_DETECTION, pAd);
			}
#endif // DFS_DEBUG //

		BBP_IO_READ8_BY_REG_ID(pAd, BBP_R115, &BBPR115);

		if (pAd->CommonCfg.RadarDetect.RDDurRegion == FCC) 
		{
#ifdef DFS_FCC_BW40_FIX
			if  ( (pAd->CommonCfg.McuRadarDetectCount == (pAd->CommonCfg.RadarDetect.DfsSessionTime + pAd->CommonCfg.RadarDetect.DfsSessionFccOff - 1)) )
#else
			if  ((pAd->CommonCfg.McuRadarDetectCount % 2) == 0)
#endif // DFS_FCC_BW40_FIX //
			{
				UCHAR bbp;
			

#ifdef DFS_DEBUG
				if (pAd->CommonCfg.McuRadarDebug & RADAR_SIMULATE2) 
				{
					if (pAd->CommonCfg.McuRadarCmd & RADAR_DETECTION)
					{
						RTMP_IO_WRITE32(pAd, 0x7100, 0x400);
						pAd->CommonCfg.McuRadarEvent |= RADAR_EVENT_WIDTH_RADAR;
					}
					pAd->CommonCfg.McuRadarDebug &= ~RADAR_SIMULATE2;

				}
#endif


				if (BBPR115 & 0x2)
				{
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R115, 0);
#ifdef DFS_DEBUG
					if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
					{
						DFS_Debug(DFS_DEBUG_LONG_REPORT, pAd);
					}
#endif // DFS_DEBUG //

					Value = BBPR115 << 24;
					BBP_IO_READ8_BY_REG_ID(pAd, BBP_R116, &bbp);
					Value |= (bbp << 16);
					BBP_IO_READ8_BY_REG_ID(pAd, BBP_R117, &bbp);
					Value |= (bbp << 8);
					BBP_IO_READ8_BY_REG_ID(pAd, BBP_R118, &bbp);
					Value |= bbp;
#ifdef DFS_FCC_BW40_FIX
					RTMP_IO_WRITE32(pAd, 0x7100, Value);
#else
					RTMP_IO_WRITE32(pAd, 0x7100 + (4 * (pAd->CommonCfg.McuRadarDetectCount / 2 - 1)) , Value);
#endif				
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x01);
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x21);
				
					pAd->CommonCfg.McuRadarEvent |= RADAR_EVENT_WIDTH_RADAR;
				}
				else
				{
					RTMP_IO_WRITE32(pAd, 0x7100 + (4 * (pAd->CommonCfg.McuRadarDetectCount / 2 - 1)) , 0);
				}
				
				
				
			}

#ifdef DFS_FCC_BW40_FIX
			if (pAd->CommonCfg.HtCapability.HtCapInfo.ChannelWidth  == BW_40)
			{
				if (pAd->CommonCfg.McuRadarDetectCount == ((pAd->CommonCfg.RadarDetect.DfsSessionTime + pAd->CommonCfg.RadarDetect.DfsSessionFccOff) - 1))
				{
				   	ULONG		Value, byteValue = 0;

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(DFS_DEBUG_SWITCH_EXT, pAd);
			}
#endif

					// Switch to Extend Channel
					if (pAd->CommonCfg.AddHTInfo.AddHtInfo.ExtChanOffset == EXTCHA_BELOW)
					{
						//  RX : control channel move to lower 
						RTMP_BBP_IO_READ8_BY_REG_ID(pAd, 3, &byteValue);
						byteValue &= (~0x20);
						RTMP_BBP_IO_WRITE8_BY_REG_ID(pAd, 3, byteValue);
					}
					else
					{
						//  RX : control channel move to upper 
						RTMP_BBP_IO_READ8_BY_REG_ID(pAd, 3, &byteValue);
						byteValue |= (0x20);
						RTMP_BBP_IO_WRITE8_BY_REG_ID(pAd, 3, byteValue);
					}

					//DisableNormalTx
					RTMP_IO_READ32(pAd, MAC_SYS_CTRL, &Value);
					RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, 0); // Toggle Tx Rx
					RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, 0xc); // Toggle Tx Rx
					RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, Value);
				}
			}
#endif // DFS_FCC_BW40_FIX //
		}
		else if ((pAd->CommonCfg.RadarDetect.RDDurRegion == JAP_W56) && ((pAd->CommonCfg.McuRadarDetectCount % 2) == 0))
		{
			UCHAR bbp;
			
			if (BBPR115 & 0x2)
			{

					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R115, 0);

#ifdef DFS_DEBUG
					if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
					{
						DFS_Debug(DFS_DEBUG_LONG_REPORT, pAd);
					}
#endif // DFS_DEBUG //

				Value = BBPR115 << 24;
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R116, &bbp);
				Value |= (bbp << 16);
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R117, &bbp);
				Value |= (bbp << 8);
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R118, &bbp);
				Value |= bbp;
				RTMP_IO_WRITE32(pAd, 0x7100 + (4 * (pAd->CommonCfg.McuRadarDetectCount / 2 - 1)) , Value);
			
				BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x01);
				BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x21);
			
				pAd->CommonCfg.McuRadarEvent |= RADAR_EVENT_WIDTH_RADAR;
				
			}
			else
			{
				RTMP_IO_WRITE32(pAd, 0x7100 + (4 * (pAd->CommonCfg.McuRadarDetectCount / 2 - 1)) , 0);
			}
						
		}
		
		if (pAd->CommonCfg.McuRadarDetectCount >= (pAd->CommonCfg.McuRadarDetectPeriod - 1))
		{

#ifdef DFS_FCC_BW40_FIX

			// Go Back to Primary Channel
			if (pAd->CommonCfg.HtCapability.HtCapInfo.ChannelWidth  == BW_40)
			{
				ULONG		Value, byteValue = 0;

				if (pAd->CommonCfg.AddHTInfo.AddHtInfo.ExtChanOffset == EXTCHA_BELOW)
				{
					//  RX : control channel move to upper 
					RTMP_BBP_IO_READ8_BY_REG_ID(pAd, 3, &byteValue);
					byteValue |= (0x20);
					RTMP_BBP_IO_WRITE8_BY_REG_ID(pAd, 3, byteValue);
				}
				else
				{
					//  RX : control channel move to lower 
					RTMP_BBP_IO_READ8_BY_REG_ID(pAd, 3, &byteValue);
					byteValue &= (~0x20);
					RTMP_BBP_IO_WRITE8_BY_REG_ID(pAd, 3, byteValue);
				}


#ifdef DFS_DEBUG
				if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
				{
					DFS_Debug(DFS_DEBUG_SWITCH_PRI, pAd);
				}
#endif // DFS_DEBUG //

			}		
#endif // DFS_FCC_BW40_FIX


			if (pAd->CommonCfg.McuRadarEvent & RADAR_EVENT_WIDTH_RADAR)
			{
				pAd->CommonCfg.McuRadarEvent &= ~(RADAR_EVENT_WIDTH_RADAR);
				RadarSMDetect(pAd, RADAR_WIDTH);
			}
			
#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_SIMULATE) 
			{
				if (pAd->CommonCfg.McuRadarCmd & RADAR_DETECTION)
				{
					RadarSMDetect(pAd, RADAR_PULSE);
				}
			}
#endif // DFS_DEBUG //
			

			if (BBPR115 & 0x1)
			{
				BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R115, 0);
#ifdef DFS_DEBUG
				if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
				{
					DFS_Debug(DFS_DEBUG_SHORT_REPORT, pAd);
				}
#endif // DFS_DEBUG //
				// pulse radar detected
				if (pAd->CommonCfg.McuRadarCmd & RADAR_DETECTION)
				{
					RadarSMDetect(pAd, RADAR_PULSE);
				}
			}


			pAd->CommonCfg.McuRadarState = FREE_FOR_TX;
			pAd->CommonCfg.McuRadarEvent &= ~(RADAR_EVENT_RADAR_DETECTING);

			RTMP_IO_WRITE32(pAd, TX_RTS_CFG, pAd->CommonCfg.OldRtsRetryLimit);
			// Restore BBP
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R65, pAd->CommonCfg._R65);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, pAd->CommonCfg._R66);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x01);

#ifdef DFS_DEBUG
			if ((pAd->CommonCfg.McuRadarDebug & RADAR_GPIO_DEBUG))
				_RTMP_IO_WRITE32(0xa0300630, 0x2000);
#endif // DFS_DEBUG //
			// EnableNormalTx
			RTMP_IO_READ32(pAd, MAC_SYS_CTRL, &Value);
			Value |= 0x04;
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, Value);

		}
		pAd->CommonCfg.McuRadarDetectCount++;

	}
	else // FREE_FOR_TX
	{

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(DFS_DEBUG_FREE_TX, pAd);
			}
#endif // DFS_DEBUG //

		if (pAd->CommonCfg.McuRadarEvent & RADAR_EVENT_CTS_SENT)
		{

#ifdef DFS_DEBUG
			if (pAd->CommonCfg.McuRadarDebug & RADAR_LOG)
			{
				DFS_Debug(DFS_DEBUG_0001, pAd);
			}
#endif // DFS_DEBUG //

			pAd->CommonCfg.McuRadarEvent &= ~RADAR_EVENT_CTS_SENT;
			pAd->CommonCfg.McuRadarDetectCount = 1;

			RTMP_IO_WRITE32(pAd, TX_RTS_CFG, pAd->CommonCfg.OldRtsRetryLimit);
			// Restore BBP
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R65, pAd->CommonCfg._R65);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, pAd->CommonCfg._R66);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x01);


#ifdef DFS_DEBUG
			if ((pAd->CommonCfg.McuRadarDebug & RADAR_GPIO_DEBUG))
				_RTMP_IO_WRITE32(0xa0300630, 0x2000);
#endif // DFS_DEBUG //
			// EnableNormalTx
			RTMP_IO_READ32(pAd, MAC_SYS_CTRL, &Value);
			Value |= 0x04;
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, Value);
		}
	}	
}
#endif // DFS_SUPPORT //
#endif // CONFIG_AP_SUPPORT //

static inline void TimerCB_Carrier(PRTMP_ADAPTER pAd)
{
	ULONG Value;

	if (pAd->CommonCfg.McuRadarEvent & RADAR_EVENT_RADAR_DETECTING)
	{
		pAd->CommonCfg.McuCarrierTick++;
		return;
	}
	else
		pAd->CommonCfg.McuCarrierTick++;


	if (pAd->CommonCfg.McuCarrierTick >= pAd->CommonCfg.McuCarrierPeriod)
	{
		// Start of period
		pAd->CommonCfg.McuCarrierTick = 0;
		pAd->CommonCfg.McuCarrierDetectCount = 0;
		pAd->CommonCfg.McuCarrierState = WAIT_CTS_BEING_SENT;
		pAd->CommonCfg.McuRadarEvent |= RADAR_EVENT_CARRIER_DETECTING;

		if ((pAd->CommonCfg.McuCarrierCtsProtect == 1)
#ifdef CARRIER_DETECTION_SUPPORT
			|| (isCarrierDetectExist(pAd))
#endif // CARRIER_DETECTION_SUPPORT //
			)
		{
			// no CTS protect
			pAd->CommonCfg.McuCarrierState = DO_DETECTION;
			MCURadarDetect(pAd);
		}
		else if (pAd->CommonCfg.McuCarrierCtsProtect == 2)
		{
			// kick CTS
			RTMP_IO_WRITE32(pAd, PBF_CTRL, 0x80);
		}
		return;
	}

	if (pAd->CommonCfg.McuCarrierState == WAIT_CTS_BEING_SENT)
	{
		// check event or timeout
		if (pAd->CommonCfg.McuRadarEvent & RADAR_EVENT_CTS_CARRIER_SENT)
		{
			pAd->CommonCfg.McuRadarEvent &= ~RADAR_EVENT_CTS_CARRIER_SENT;
			pAd->CommonCfg.McuCarrierDetectCount = 1;
			pAd->CommonCfg.McuCarrierState = DO_DETECTION;

			return;
		}
		
		if (pAd->CommonCfg.McuCarrierDetectCount >= (5-1)) // timeout 5ms, start from 0
		{
			pAd->CommonCfg.McuCarrierState = FREE_FOR_TX; // timeout 5ms, give up
			pAd->CommonCfg.McuRadarEvent &= ~(RADAR_EVENT_CARRIER_DETECTING);

		}
		
		pAd->CommonCfg.McuCarrierDetectCount++;
	}
	else if (pAd->CommonCfg.McuCarrierState == DO_DETECTION)
	{
		if (pAd->CommonCfg.McuCarrierDetectCount >= (pAd->CommonCfg.McuCarrierDetectPeriod - 1))
		{
			UCHAR BBPR115;

			BBP_IO_READ8_BY_REG_ID(pAd, BBP_R115, &BBPR115);
	#ifdef CONFIG_AP_SUPPORT
			CarrierDetectionCheck(pAd);
	#endif
			pAd->CommonCfg.McuCarrierState = FREE_FOR_TX;
			pAd->CommonCfg.McuRadarEvent &= ~(RADAR_EVENT_CARRIER_DETECTING);

			RTMP_IO_WRITE32(pAd, TX_RTS_CFG, pAd->CommonCfg.OldRtsRetryLimit);
			// Restore BBP

			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, pAd->CommonCfg._R66);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R69, pAd->CommonCfg._R69);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R70, pAd->CommonCfg._R70);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R73, pAd->CommonCfg._R73);

#ifdef DFS_DEBUG
			if ((pAd->CommonCfg.McuRadarDebug & RADAR_GPIO_DEBUG))
				_RTMP_IO_WRITE32(0xa0300630, 0x2000);
#endif // DFS_DEBUG //
			// EnableNormalTx
			RTMP_IO_READ32(pAd, MAC_SYS_CTRL, &Value);
			Value |= 0x04;
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, Value);

			

		}
		pAd->CommonCfg.McuCarrierDetectCount++;
	}
	else // FREE_FOR_TX
	{
		if (pAd->CommonCfg.McuRadarEvent & RADAR_EVENT_CTS_CARRIER_SENT)
		{
			pAd->CommonCfg.McuRadarEvent &= ~RADAR_EVENT_CTS_CARRIER_SENT;
			pAd->CommonCfg.McuCarrierDetectCount = 1;

			RTMP_IO_WRITE32(pAd, TX_RTS_CFG, pAd->CommonCfg.OldRtsRetryLimit);
			// Restore BBP
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, pAd->CommonCfg._R66);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R69, pAd->CommonCfg._R69);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R70, pAd->CommonCfg._R70);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R73, pAd->CommonCfg._R73);


#ifdef DFS_DEBUG
			if ((pAd->CommonCfg.McuRadarDebug & RADAR_GPIO_DEBUG))
				_RTMP_IO_WRITE32(0xa0300630, 0x2000);
#endif // DFS_DEBUG //
			// EnableNormalTx
			RTMP_IO_READ32(pAd, MAC_SYS_CTRL, &Value);
			Value |= 0x04;
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, Value);
		}
	}	
}


void MCURadarDetect(PRTMP_ADAPTER pAd)
{
	ULONG Value;
	UCHAR bbp;
	
			BBP_IO_READ8_BY_REG_ID(pAd, BBP_R113, &bbp);
			if (bbp & 0x20)
				return;

			//DisableNormalTx
			RTMP_IO_READ32(pAd, MAC_SYS_CTRL, &Value);
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, 0);
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, 0xc);
			Value &= (ULONG)~(0x04);
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, Value);
						
#ifdef DFS_DEBUG
			if ((pAd->CommonCfg.McuRadarDebug & RADAR_GPIO_DEBUG))
				if (pAd->CommonCfg.McuRadarCtsProtect != 1)
					_RTMP_IO_WRITE32(0xa030062c, 0x2000);
#endif // DFS_DEBUG //


			RTMP_IO_READ32(pAd, TX_RTS_CFG, &pAd->CommonCfg.OldRtsRetryLimit);
			RTMP_IO_WRITE32(pAd, TX_RTS_CFG, 0);
			
			//change BBP now
			if (pAd->CommonCfg.McuRadarEvent & RADAR_EVENT_CARRIER_DETECTING)
			{
				UCHAR BBPR4;
				
				// Clear False CCA
				RTMP_IO_READ32(pAd, RX_STA_CNT1, &Value);
				
				// save old value
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R66, &pAd->CommonCfg._R66);
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R69, &pAd->CommonCfg._R69);
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R70, &pAd->CommonCfg._R70);
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R73, &pAd->CommonCfg._R73);

				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R4, &BBPR4);
				if (BBPR4 & 0x18) // BW 40
				{
					if (pAd->CommonCfg.Channel > 14)
					{
						// BW 40, A band
						BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, 0x50);
					}
					else
					{
						// BW 40, G band
						BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, 0x40);
						AsicDisableSync(pAd);
					}
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R69, 0x30);
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R70, 0x30);
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R73, 0x08);
				}
				else // BW 20
				{
					if (pAd->CommonCfg.Channel > 14)
					{
						// BW 20, A band
						BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, 0x60);
					}
					else
					{
						// BW 20, G band
						BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, 0x30);
					}
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R69, 0x28);
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R70, 0x28);
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R73, 0x08);
				}
				
			}
			else // RADAR_DETECTION
			{
				// save old value				
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R65, &pAd->CommonCfg._R65);
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R66, &pAd->CommonCfg._R66);

				BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R65, pAd->CommonCfg.R65);
				BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, pAd->CommonCfg.R66);
				BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x21);
			}		
			return;
}

#ifdef CONFIG_RALINK_RT3052
#ifdef CARRIER_DETECTION_SUPPORT

// for debug and test
static ULONG time[20];
static ULONG idle[20];
static ULONG busy[20];
static ULONG cd_idx=0;

static void ToneRadarProgram(PRTMP_ADAPTER pAd);
static void ToneRadarEnable(PRTMP_ADAPTER pAd);

void RTMPHandleRadarInterrupt(PRTMP_ADAPTER  pAd)
{
	UINT32 value, delta;
	
	RTMP_IO_READ32(pAd, PBF_LIFE_TIMER, &value);
	RTMP_IO_READ32(pAd, CH_IDLE_STA, &pAd->CommonCfg.CarrierDetect.idle_time);
	RTMP_IO_READ32(pAd, CH_BUSY_STA, &pAd->CommonCfg.CarrierDetect.busy_time);
	delta = (value >> 4) - pAd->CommonCfg.CarrierDetect.TimeStamp;
	pAd->CommonCfg.CarrierDetect.TimeStamp = value >> 4;
	pAd->CommonCfg.CarrierDetect.OneSecIntCount++;
	
	if (pAd->CommonCfg.CarrierDetect.Debug)
	{
		if (cd_idx < 20)
		{
			time[cd_idx] = delta;
			idle[cd_idx] = pAd->CommonCfg.CarrierDetect.idle_time;
			busy[cd_idx] = pAd->CommonCfg.CarrierDetect.busy_time;
			cd_idx++;
		}
		else
		{
			int i;
			pAd->CommonCfg.CarrierDetect.Debug = 0;
			for (i = 0; i < 20; i++)
			{
				printk("%3d %4d %d %d\n", i, time[i], idle[i], busy[i]);
			}
			cd_idx = 0;
			
		}
	}
	

	if (pAd->CommonCfg.CarrierDetect.CD_State == CD_NORMAL)
	{

		if ((delta < pAd->CommonCfg.CarrierDetect.criteria) && (pAd->CommonCfg.CarrierDetect.recheck))
			pAd->CommonCfg.CarrierDetect.recheck --;
		else
			pAd->CommonCfg.CarrierDetect.recheck = pAd->CommonCfg.CarrierDetect.recheck1;

		if (pAd->CommonCfg.CarrierDetect.recheck == 0)
		{
			// declare carrier sense
			pAd->CommonCfg.CarrierDetect.CD_State = CD_SILENCE;
			//pAd->CommonCfg.CarrierDetect.recheck = pAd->CommonCfg.CarrierDetect.recheck2;
			pAd->CommonCfg.CarrierDetect.recheck = pAd->CommonCfg.CarrierDetect.recheck1;
			if (pAd->CommonCfg.CarrierDetect.CarrierDebug == 0)
			{
	
				DBGPRINT(RT_DEBUG_TRACE, ("Carrier Detected\n"));

				// disconnect all STAs behind AP.
				//MacTableReset(pAd);
				
				// stop all TX actions including Beacon sending.
				AsicDisableSync(pAd);
			}
			else
			{
				printk("Carrier Detected\n");
			}
			

		}
	}
#if 0
	else if (pAd->CommonCfg.CarrierDetect.CD_State == CD_SILENCE)
	{
		// check if carrier gone??
		if ((delta > (pAd->CommonCfg.CarrierDetect.criteria * CARRIER_DETECT_STOP_RATIO)) && (pAd->CommonCfg.CarrierDetect.recheck))
			pAd->CommonCfg.CarrierDetect.recheck --;
		if (pAd->CommonCfg.CarrierDetect.recheck == 0)
		{
			pAd->CommonCfg.CarrierDetect.CD_State = CD_NORMAL;
			pAd->CommonCfg.CarrierDetect.recheck = pAd->CommonCfg.CarrierDetect.recheck1;
			if (pAd->CommonCfg.CarrierDetect.CarrierDebug == 0)
			{
				DBGPRINT(RT_DEBUG_TRACE, ("Carrier gone\n"));
				// start all TX actions.
				APMakeAllBssBeacon(pAd);
				APUpdateAllBeaconFrame(pAd);
				AsicEnableBssSync(pAd);
			}
			else
			{
				printk("Carrier gone\n");
			}
		}
		
	}
#endif	
	
	if (pAd->CommonCfg.CarrierDetect.Enable)
	{
		ToneRadarProgram(pAd);
		ToneRadarEnable(pAd);
	}
}



static void ToneRadarProgram(PRTMP_ADAPTER pAd)
{
	UCHAR bbp;
	ULONG threshold;

	// programe delta delay & division bit
	BBP_IO_WRITE8_BY_REG_ID(pAd, 184, 0xf0);
	bbp = pAd->CommonCfg.CarrierDetect.delta << 4;
	bbp |= (pAd->CommonCfg.CarrierDetect.div_flag & 0x1) << 3;
	BBP_IO_WRITE8_BY_REG_ID(pAd, 185, bbp);
	
	// if wireless mode is 20Mhz mode, then the threshold should div by 2
	if (pAd->CommonCfg.HtCapability.HtCapInfo.ChannelWidth  == BW_20)
	{
		threshold = pAd->CommonCfg.CarrierDetect.threshold >> 1;
	}
	else
		threshold = pAd->CommonCfg.CarrierDetect.threshold;
	
	// program threshold
	BBP_IO_WRITE8_BY_REG_ID(pAd, 184, 0x34);
	BBP_IO_WRITE8_BY_REG_ID(pAd, 185, (threshold & 0xff000000) >> 24);
	
	BBP_IO_WRITE8_BY_REG_ID(pAd, 184, 0x24);
	BBP_IO_WRITE8_BY_REG_ID(pAd, 185, (threshold & 0xff0000) >> 16);
	
	BBP_IO_WRITE8_BY_REG_ID(pAd, 184, 0x14);
	BBP_IO_WRITE8_BY_REG_ID(pAd, 185, (threshold & 0xff00) >> 8);

	BBP_IO_WRITE8_BY_REG_ID(pAd, 184, 0x04);
	BBP_IO_WRITE8_BY_REG_ID(pAd, 185, threshold & 0xff);
}

static void ToneRadarEnable(PRTMP_ADAPTER pAd)
{
	BBP_IO_WRITE8_BY_REG_ID(pAd, 184, 0x05);
}


void NewCarrierDetectionStart(PRTMP_ADAPTER pAd)
{	
	ULONG Value;
	// Enable Bandwidth usage monitor
		
	RTMP_IO_READ32(pAd, CH_TIME_CFG, &Value);
	RTMP_IO_WRITE32(pAd, CH_TIME_CFG, Value | 0x1f);
	pAd->CommonCfg.CarrierDetect.recheck1 = CARRIER_DETECT_RECHECK_TIME;
	pAd->CommonCfg.CarrierDetect.recheck2 = CARRIER_DETECT_STOP_RECHECK_TIME;
	

	// Init Carrier Detect
	if (pAd->CommonCfg.CarrierDetect.Enable)
	{
		pAd->CommonCfg.CarrierDetect.TimeStamp = 0;
		pAd->CommonCfg.CarrierDetect.recheck = pAd->CommonCfg.CarrierDetect.recheck1;
		ToneRadarProgram(pAd);
		ToneRadarEnable(pAd);
	}
	
}
#endif // CARRIER_DETECTION_SUPPORT //

#endif // CONFIG_RALINK_RT3052 //


#ifdef NEW_DFS

typedef struct _NewDFSProgParam
{
	UCHAR channel;
	UCHAR mode;
	USHORT avgLen; // M = Average length
	USHORT ELow;
	USHORT EHigh;
	USHORT WLow;
	USHORT WHigh;
	USHORT EpsilonW;
	ULONG TLow;
	ULONG THigh;
	USHORT EpsilonT;
}NewDFSProgParam, *pNewDFSProgParam;

#define NEW_DFS_FCC		0x1 // include Japan
#define NEW_DFS_EU		0x2
#define NEW_DFS_JAP		0x4
#define NEW_DFS_END		0xff
#define NEW_DFS_MAX_CHANNEL	4
#define MAX_VALID_RADAR_W	5
#define MAX_VALID_RADAR_T	5

typedef struct _NewDFSTable
{
	USHORT type;
	NewDFSProgParam entry[NEW_DFS_MAX_CHANNEL];
}NewDFSTable, *pNewDFSTable;



static NewDFSTable NewDFSTable1[] = 
{
	{
		NEW_DFS_FCC,
		{
		{0, 0,  20,  18,  26,   10, 2000, 1,  3000, 30000, 5},
		{1, 0, 100, 100, 140,   12, 2000, 3,  2500, 30000, 5},
		{2, 0, 100,  60, 120,   12, 2000, 3,  3000, 30000, 5},
		{3, 2, 200,  20, 190, 700, 2000,  50, 15000, 45000, 200},
		}
	},
	{
		NEW_DFS_EU,
		{
		//{0, 0, 16, 10, 15, 4, 10, 3, 4000, 85000, 100},
		//{1, 0, 32, 15, 28, 8, 100, 6, 4000, 85000, 100},
		//{2, 0, 80, 20, 75, 10, 300, 10, 20000, 85000, 250},
		//{3, 2, 250, 20, 200, 150, 600, 50, 30000, 90000, 250},
		{0, 0, 20,	18,	26,	10,	2000,	1, 4900, 101000, 5},
		{1, 0, 70,	100,140, 12, 3000,	3, 4900, 101000, 5},
		{2, 0, 80,	60, 120, 12, 3000,	3, 4900, 101000, 5},
		{3, 2, 250,	20, 200, 150, 600,	50, 30000, 90000, 250},
		}
	},
	{
		NEW_DFS_JAP,
		{
		{0, 0,  10,  10,  15,   3,  2000,  1,  3000, 85000, 10},
		{1, 0, 70, 100, 140,   12, 2000, 3,  2500, 85000, 10},
		{2, 0, 80,  60, 120,   12, 2000, 3,  3000, 85000, 10},
		{3, 2, 200,  20, 190, 700, 2000,  50, 15000, 45000, 200},
		}
	}
	
};

typedef struct _NewDFSValidRadar
{
	USHORT type;
	USHORT channel; // bit map
	USHORT WLow;
	USHORT WHigh;
	USHORT W[MAX_VALID_RADAR_W];
	USHORT WMargin;
	ULONG TLow;
	ULONG THigh;
	ULONG T[MAX_VALID_RADAR_T];
	USHORT TMargin;
}NewDFSValidRadar, *pNewDFSValidRadar;

#ifdef DFS_DEBUG


typedef struct _NewDFSDebugResult
{
	char delta_delay_shift;
	char EL_shift;
	char EH_shift;
	char WL_shift;
	char WH_shift;
	ULONG hit_time;
	ULONG false_time;
}NewDFSDebugResult, *pNewDFSDebugResult;

NewDFSDebugResult TestResult[1000];

#endif


NewDFSValidRadar NewDFSValidTable[] = 
{
	// FCC-1  && (Japan W53 Radar 1 / W56 Radar 2)
	{
	(NEW_DFS_FCC | NEW_DFS_JAP),
	//3,
	7,
	//0, 0, 
	//{20, 0, 0, 0, 0},
	10, 1000,
	{0, 0, 0, 0, 0},
	4,
	0, 0, 
	{28560 - 70, 0, 0, 0, 0},
	150
	},
	// FCC-2
	{
	(NEW_DFS_FCC | NEW_DFS_JAP),
	//3,
	7,
	//20, 100, 
	13, 1000,
	{0, 0, 0, 0, 0},
	1,
	3000, 4600 - 20,
	{0, 0, 0, 0, 0},
	25
	},
	// FCC-3 & FCC-4
	{
	(NEW_DFS_FCC | NEW_DFS_JAP),
	//6,
	7,
	//120, 200, FCC-3 
	//220, 400, FCC-4
	100, 1500, 
	{0, 0, 0, 0, 0},
	1,
	4000, 10000 - 40, 
	{0, 0, 0, 0, 0},
	60
	},
#if 0 // detect with another method
	//FCC-5
	{
	NEW_DFS_FCC,
	8,
	100, 2000, 
	{0, 0, 0, 0, 0},
	50,
	20000, 40000, 
	{0, 0, 0, 0, 0},
	150
	},
#endif
	// FCC-6
	{
	(NEW_DFS_FCC | NEW_DFS_JAP),
	//3,
	7,
	//0, 0, 
	//{20, 0, 0, 0, 0},
	12, 1000,
	{0, 0, 0, 0, 0},
	1,
	0, 0, 
	{6660-10, 0, 0, 0, 0},
	35
	},
	// Japan W53 Radar 2
	{
	NEW_DFS_JAP,
	7,
	//0, 0, 
	//{50, 0, 0, 0, 0},
	40, 1000, 
	{0, 0, 0, 0, 0},
	1,
	0, 0, 
	{76923 - 30, 0, 0, 0, 0},
	180
	},
	// Japan W56 Radar 1
	{
	NEW_DFS_JAP,
	7,
	//0, 0, 
	//{10, 0, 0, 0, 0},
	5, 500, 
	{0, 0, 0, 0, 0},
	2,
	0, 0, 
	{27777 - 30, 0, 0, 0, 0},
	70
	},
	// Japan W56 Radar 3
	{
	NEW_DFS_JAP,
	7,
	//0, 0, 
	//{40, 0, 0, 0, 0},
	30, 1000, 
	{0, 0, 0, 0, 0},
	1,
	0, 0, 
	{80000 - 50, 0, 0, 0, 0},
	200
	},
	// EN-1
	{
	NEW_DFS_EU,
	3,
	//0, 0, 
	//{20, 0, 0, 0, 0},
	12, 1000, 
	{0, 0, 0, 0, 0},
	1,
	0, 0, 
	//{15000, 0, 0, 0, 0},
	{26666 - 20, 0, 0, 0, 0},
	100
	},
	// EN-2 / EN-3
	{
	NEW_DFS_EU,
	7,
	//0, 0, 
	//{20, 40, 100, 200, 300},
	12, 2000, 
	{0, 0, 0, 0, 0},
	1,
	0, 0, 
	//{4000, 6000, 10000, 16000, 20000},
	{20000 - 20, 25000 - 25, 40000 - 35, 66666 - 50, 100000 - 80},
	200
	},
	// EN-4
	{
	NEW_DFS_EU,
	7,
	//0, 0, 
	//{20, 40, 100, 200, 300},
	12, 2000, 
	{0, 0, 0, 0, 0},
	1,
	0, 0, 
	//{24000, 30000, 32000, 0, 0},
	{16666 - 13, 13333 - 10, 12500 -10, 0, 0},
	100
	},
	// EN-5
	{
	NEW_DFS_EU,
	7,
	//0, 0, 
	//{20, 40, 100, 200, 300},
	12, 2000, 
	{0, 0, 0, 0, 0},
	1,
	0, 0, 
	//{46000, 60000, 70000, 80000, 0},
	{8695 - 7, 6666 - 5, 5714 - 4, 5000, 0},
	50
	},
	// EN-6
	{
	NEW_DFS_EU,
	8,
	//0, 0, 
	//{400, 600, 0, 0, 0},
	380, 3000, 
	{0, 0, 0, 0, 0},
	10,
	0, 0, 
	//{40000, 60000, 80000, 0, 0},
	{10000 - 10, 6666 - 5, 5000 - 4, 0, 0},
	40
	},
	
	{
	NEW_DFS_END,
	0,
	0, 0, 
	{0, 0, 0, 0, 0},
	0,
	0, 0, 
	{0, 0, 0, 0, 0},
	0,
	},
};

void 	modify_table1(PRTMP_ADAPTER pAd, ULONG idx, ULONG value)
{
	pNewDFSTable pDFS2Table;
	ULONG x, y;	
	
	if (idx <= 40)
	{
		if (pAd->CommonCfg.MCURadarRegion == NEW_DFS_FCC)
			pDFS2Table = &NewDFSTable1[0];
		else if (pAd->CommonCfg.MCURadarRegion == NEW_DFS_EU)
			pDFS2Table = &NewDFSTable1[1];
		else // Japan
			pDFS2Table = &NewDFSTable1[2];
	}
	else
		return;

	idx--;
	
	x = idx / 10;
	y = idx % 10;

	switch (y)
	{
	case 0:
		pDFS2Table->entry[x].mode = (USHORT)value;
		break;
	case 1:
		pDFS2Table->entry[x].avgLen = (USHORT)value;
		break;
	case 2:
		pDFS2Table->entry[x].ELow = (USHORT)value;
		break;

	case 3:
		pDFS2Table->entry[x].EHigh = (USHORT)value;
		break;

	case 4:
		pDFS2Table->entry[x].WLow = (USHORT)value;
		break;

	case 5:
		pDFS2Table->entry[x].WHigh = (USHORT)value;
		break;

	case 6:
		pDFS2Table->entry[x].EpsilonW = (USHORT)value;
		break;

	case 7:
		pDFS2Table->entry[x].TLow = (ULONG)value;
		break;

	case 8:
		pDFS2Table->entry[x].THigh = (ULONG)value;
		break;

	case 9:
		pDFS2Table->entry[x].EpsilonT = (USHORT)value;
		break;
	default:
		break;
	}

	for (x = 0; x < 4; x++)
		printk("mode %d, (M)%03d  (E)%02d %03d (W) %03d %04d <EW> %02d (T) %05d %05d <ET> %03d\n", pDFS2Table->entry[x].mode, pDFS2Table->entry[x].avgLen,
		pDFS2Table->entry[x].ELow, pDFS2Table->entry[x].EHigh, pDFS2Table->entry[x].WLow, pDFS2Table->entry[x].WHigh,
		pDFS2Table->entry[x].EpsilonW, pDFS2Table->entry[x].TLow, pDFS2Table->entry[x].THigh, pDFS2Table->entry[x].EpsilonT);
	
}



void 	modify_table2(PRTMP_ADAPTER pAd, ULONG idx, ULONG value)
{
	pNewDFSValidRadar pDFSValidRadar;
	ULONG x, y, z;
	
	idx--;

	x = idx / 17;
	y = idx % 17;
	
	pDFSValidRadar = &NewDFSValidTable[0];
	
	while (pDFSValidRadar->type != NEW_DFS_END)
	{
		if (pDFSValidRadar->type & pAd->CommonCfg.MCURadarRegion)
		{
			if (x == 0)
				break;
			else
			{
				x--;
				pDFSValidRadar++;
			}
		}
		else
			pDFSValidRadar++;
	}
	
	if (pDFSValidRadar->type == NEW_DFS_END)
	{
		printk("idx=%d exceed max number\n", idx);
		return;
	}
	switch(y)
	{
	case 0:
		pDFSValidRadar->channel = value;
		break;
	case 1:
		pDFSValidRadar->WLow = value;
		break;
	case 2:
		pDFSValidRadar->WHigh = value;
		break;
	case 3:
		pDFSValidRadar->W[0] = value;
		break;
	case 4:
		pDFSValidRadar->W[1] = value;
		break;
	case 5:
		pDFSValidRadar->W[2] = value;
		break;
	case 6:
		pDFSValidRadar->W[3] = value;
		break;
	case 7:
		pDFSValidRadar->W[5] = value;
		break;
	case 8:
		pDFSValidRadar->WMargin = value;
		break;
	case 9:
		pDFSValidRadar->TLow = value;
		break;
	case 10:
		pDFSValidRadar->THigh = value;
		break;
	case 11:
		pDFSValidRadar->T[0] = value;
		break;
	case 12:
		pDFSValidRadar->T[1] = value;
		break;
	case 13:
		pDFSValidRadar->T[2] = value;
		break;
	case 14:
		pDFSValidRadar->T[3] = value;
		break;
	case 15:
		pDFSValidRadar->T[4] = value;
		break;
	case 16:
		pDFSValidRadar->TMargin = value;
		break;
	}
	
	
	USHORT type;
	
	USHORT channel; // bit map
	USHORT WLow;
	USHORT WHigh;
	USHORT W[MAX_VALID_RADAR_W];
	USHORT WMargin;
	ULONG TLow;
	ULONG THigh;
	ULONG T[MAX_VALID_RADAR_T];
	USHORT TMargin;

	pDFSValidRadar = &NewDFSValidTable[0];
	while (pDFSValidRadar->type != NEW_DFS_END)
	{
		if (pDFSValidRadar->type & pAd->CommonCfg.MCURadarRegion)
		{
			printk("ch = %x  --- ", pDFSValidRadar->channel);
			printk("wl:wh = %d:%d  ", pDFSValidRadar->WLow, pDFSValidRadar->WHigh);
			printk("w[] = %d %d %d %d %d --- ", pDFSValidRadar->W[0], pDFSValidRadar->W[1], pDFSValidRadar->W[2], pDFSValidRadar->W[3], pDFSValidRadar->W[4]);
			printk("W Margin = %d\n", pDFSValidRadar->WMargin);
			printk("        --- Tl:Th = %d:%d  ", pDFSValidRadar->TLow, pDFSValidRadar->THigh);
			printk("T[] = %d %d %d %d %d --- ", pDFSValidRadar->T[0], pDFSValidRadar->T[1], pDFSValidRadar->T[2], pDFSValidRadar->T[3], pDFSValidRadar->T[4]);
			printk("T Margin = %d\n", pDFSValidRadar->TMargin);
		}
		pDFSValidRadar++;
	}

}


#define NEW_DFS_CHECK_TIME			300
#define NEW_DFS_WATCH_DOG_TIME		1 // note that carrier detection also need timer interrupt hook

static VOID NewTimerCB_Radar(PRTMP_ADAPTER pAd)
{
	UCHAR BBPR141;
	pNewDFSTable pDFS2Table;
	pNewDFSValidRadar pDFSValidRadar;
	UCHAR channel, channelBit;
	UCHAR radarDeclared;
	INT i, j;
	UCHAR BBPR140;
	UCHAR BBPR156, BBPR157, BBPR158, BBPR159;
	UCHAR BBPR160, BBPR161, BBPR162, BBPR163;
	USHORT NewDfsCheckTime = NEW_DFS_CHECK_TIME;
	

	if (pAd->CommonCfg.PollTime != 0)
		NewDfsCheckTime = pAd->CommonCfg.PollTime;
	
	if (pAd->CommonCfg.McuRadarTick++ > NewDfsCheckTime)
	{
		radarDeclared = 0;
		pAd->CommonCfg.McuRadarTick = 0;

		{

			// Bandwidth usage monitor
			//pAd->CommonCfg.rssi = RTMPMaxRssi(pAd, (CHAR)pAd->ApCfg.RssiSample.AvgRssi0,
			//				(CHAR)pAd->ApCfg.RssiSample.AvgRssi1,
			//				(CHAR)pAd->ApCfg.RssiSample.AvgRssi2);
			RTMP_IO_READ32(pAd, CH_IDLE_STA, &pAd->CommonCfg.idle_time);
			RTMP_IO_READ32(pAd, CH_BUSY_STA, &pAd->CommonCfg.busy_time);

			pAd->CommonCfg.rssi = RTMPMaxRssi(pAd, (CHAR)pAd->ApCfg.RssiSample.AvgRssi0,
						(CHAR)pAd->ApCfg.RssiSample.AvgRssi1,
						(CHAR)pAd->ApCfg.RssiSample.AvgRssi2);


			pAd->CommonCfg.ch_busy = 0;
			if (pAd->CommonCfg.rssi != 0)
			{
				// The False Detection Filter is :
				// if rssi is large than a certain value and channel bandwidth usage is more than a certain percentage
				// then any radar detected count as false detection.
				for (i = 0; i < pAd->CommonCfg.fdf_num; i++)
				{
					if (((pAd->CommonCfg.ch_busy_threshold[i] * pAd->CommonCfg.idle_time) < (pAd->CommonCfg.busy_time * (100 - pAd->CommonCfg.ch_busy_threshold[i]))) 
					 && (pAd->CommonCfg.rssi > pAd->CommonCfg.rssi_threshold[i]))
						pAd->CommonCfg.ch_busy = 1;
				}
			}
			
#ifdef DFS_DEBUG
			DBGPRINT(RT_DEBUG_INFO, ("%d:%d,%d (%d)\n", pAd->CommonCfg.idle_time, pAd->CommonCfg.busy_time, pAd->CommonCfg.rssi, pAd->CommonCfg.ch_busy));
#endif
			// Read debug port
			if (pAd->CommonCfg.RadarDetect.RDDurRegion != CE)
			{
				int k;
				UCHAR BBPR127;
				ULONG time;
				USHORT width;
				UCHAR id;
				
				// disable debug mode to read debug port of channel 3
				BBP_IO_WRITE8_BY_REG_ID(pAd, 126, 2);
				
				pAd->CommonCfg.fcc_5_counter++;

				for (k = 0; k < 384; k++)
				{
					BBPR127 = 0;
					BBP_IO_READ8_BY_REG_ID(pAd, 127, &BBPR127);
					//printk ("%02x ", BBPR127);
					// print current time for ref
					
					switch(k % 6)
					{
					case 0:
						width = 0;
						time = 0;
						id = BBPR127;
						break;
					case 1:
						time |= BBPR127;
						break;
					case 2:
						time |= BBPR127 << 8;
						break;
					case 3:
						time |= (BBPR127 & 0x3f) << 16;
						width |= (BBPR127 & 0xc0) >> 6;
						break;
					case 4:
						width |= BBPR127 << 2;
						break;
					case 5:
						width |= (BBPR127 & 0x7) << 10;
						if (id == 0x3)
						{
							if (pAd->CommonCfg.NewDfsDebug & RADAR_DEBUG_EVENT)
							{
								printk("counter = %d  ", pAd->CommonCfg.fcc_5_counter);
								printk("time = %d  ", time);
								printk("width = %d\n", width);
							}
							pAd->CommonCfg.FCC_5[pAd->CommonCfg.fcc_5_idx].counter = pAd->CommonCfg.fcc_5_counter;
							pAd->CommonCfg.FCC_5[pAd->CommonCfg.fcc_5_idx].timestamp = time;
							pAd->CommonCfg.FCC_5[pAd->CommonCfg.fcc_5_idx].width = width;
							
							if (pAd->CommonCfg.fcc_5_last_idx != pAd->CommonCfg.fcc_5_idx)
							{
								int delta;
								
								if (pAd->CommonCfg.FCC_5[pAd->CommonCfg.fcc_5_idx].counter == pAd->CommonCfg.FCC_5[pAd->CommonCfg.fcc_5_last_idx].counter)
								{
									delta = pAd->CommonCfg.FCC_5[pAd->CommonCfg.fcc_5_idx].width - pAd->CommonCfg.FCC_5[pAd->CommonCfg.fcc_5_last_idx].width;

									printk("w0 = %d, timestamp = %d\n", pAd->CommonCfg.FCC_5[pAd->CommonCfg.fcc_5_last_idx].width, pAd->CommonCfg.FCC_5[pAd->CommonCfg.fcc_5_last_idx].timestamp);
									printk("w1 = %d, timestamp = %d\n", pAd->CommonCfg.FCC_5[pAd->CommonCfg.fcc_5_idx].width, pAd->CommonCfg.FCC_5[pAd->CommonCfg.fcc_5_idx].timestamp);
									// debug
									if (delta < pAd->CommonCfg.fcc_5_threshold && delta > (-1 * pAd->CommonCfg.fcc_5_threshold))
									{
										// looks like there are no long pulse false detection
										//if (pAd->CommonCfg.ch_busy == 0)
										//{
											printk("radar detected!!!\n");
											radarDeclared = 1;
										//}
#ifdef DFS_DEBUG
										//else
										//	printk("FDF: long pulse\n");
										
										DBGPRINT(RT_DEBUG_TRACE, ("%d:%d,%d (%d)\n", pAd->CommonCfg.idle_time, pAd->CommonCfg.busy_time, pAd->CommonCfg.rssi, pAd->CommonCfg.ch_busy));
#endif
									}
									
								}
							}

							pAd->CommonCfg.fcc_5_last_idx = pAd->CommonCfg.fcc_5_idx;
							pAd->CommonCfg.fcc_5_idx++;
							if (pAd->CommonCfg.fcc_5_idx >= NEW_DFS_FCC_5_ENT_NUM)
								pAd->CommonCfg.fcc_5_idx = 0;
						}

						break;
						
					}
						
				}
				BBP_IO_READ8_BY_REG_ID(pAd, 127, &BBPR127);
				BBP_IO_READ8_BY_REG_ID(pAd, 127, &BBPR127);
				BBP_IO_READ8_BY_REG_ID(pAd, 127, &BBPR127);
				BBP_IO_READ8_BY_REG_ID(pAd, 127, &BBPR127);
				BBP_IO_READ8_BY_REG_ID(pAd, 127, &BBPR127);

				// enable debug mode
				BBP_IO_WRITE8_BY_REG_ID(pAd, 126, 3);
			}
		}
		
		
		// Poll Status register
		BBP_IO_READ8_BY_REG_ID(pAd, 141, &BBPR141);
		channel = BBPR141 & 0xf;

#if 0
		if (pAd->CommonCfg.DfsDebug.start == 2)
		{
			pAd->CommonCfg.DfsDebug.count++;
			if (pAd->CommonCfg.DfsDebug.count >= pAd->CommonCfg.DfsDebug.wait_time)
			{
				pAd->CommonCfg.DfsDebug.count = 0;
				RadarDebugNext(pAd);
			}
		}
#endif		
		if (channel)
		{
			UCHAR WMatched, TMatched;
			USHORT W, W1;
			ULONG T, T1;

			DBGPRINT(RT_DEBUG_TRACE, ("NewTimerCB_Radar, BBPR141 = 0x%x, channel bit= 0x%x\n", BBPR141, channel));
			
			if (pAd->CommonCfg.HtCapability.HtCapInfo.ChannelWidth  == BW_40)
				BBPR140 = 0x80 | pAd->CommonCfg.ChEnable;
			else
				BBPR140 = 0 | pAd->CommonCfg.ChEnable;

			{
			
				for (j = 0;j < NEW_DFS_MAX_CHANNEL; j++)
				{
					if (channel & (0x1 << j))
					{

						// select Channel first
						BBP_IO_WRITE8_BY_REG_ID(pAd, 140, BBPR140 | (j << 4));
						DBGPRINT(RT_DEBUG_TRACE, ("Select Channel = 0x%x\n", BBPR140 | (j << 4)));
						BBP_IO_READ8_BY_REG_ID(pAd, 160, &BBPR160);
						BBP_IO_READ8_BY_REG_ID(pAd, 161, &BBPR161);
						BBP_IO_READ8_BY_REG_ID(pAd, 162, &BBPR162);
						BBP_IO_READ8_BY_REG_ID(pAd, 163, &BBPR163);
						//printk("... 160 ~ 163  %x %x %x %x\n", BBPR160, BBPR161, BBPR162, BBPR163);
						BBP_IO_READ8_BY_REG_ID(pAd, 156, &BBPR156);
						BBP_IO_READ8_BY_REG_ID(pAd, 157, &BBPR157);
						BBP_IO_READ8_BY_REG_ID(pAd, 158, &BBPR158);
						BBP_IO_READ8_BY_REG_ID(pAd, 159, &BBPR159);
						//printk("... 156 ~ 159  %x %x %x %x\n", BBPR156, BBPR157, BBPR158, BBPR159);


						pDFSValidRadar = &NewDFSValidTable[0];

						while (pDFSValidRadar->type != NEW_DFS_END)
						{
							if ((pDFSValidRadar->type & pAd->CommonCfg.MCURadarRegion) == 0)
							{
								pDFSValidRadar++;
								continue;
							}

							if (! (pDFSValidRadar->channel & channel))
							{
								pDFSValidRadar++;
								continue;
							}

							if (pAd->CommonCfg.MCURadarRegion == NEW_DFS_FCC)
								pDFS2Table = &NewDFSTable1[0];
							else if (pAd->CommonCfg.MCURadarRegion == NEW_DFS_EU)
								pDFS2Table = &NewDFSTable1[1];
							else // Japan
								pDFS2Table = &NewDFSTable1[2];

							
								
							//printk("Table type = %d, Channel bit map = 0x%x\n", pDFSValidRadar->type, pDFSValidRadar->channel);

							W = 0;
							T = 0;
							T |= BBPR156 << 8;
							T |= BBPR157;
							T |= (BBPR158 & 0x80) << 9;
			
							W |= (BBPR158 & 0xf) << 8;
							W |= BBPR159;

							if ( (pDFS2Table->entry[j].mode == 1) || (pDFS2Table->entry[j].mode == 2) || (pDFS2Table->entry[j].mode == 6) )
							{
								W1 = 0;
								T1 = 0;
								T1 |= BBPR160 << 8;
								T1 |= BBPR161;
								T1 |= (BBPR162 & 0x80) << 9;
			
								W1 |= (BBPR162 & 0xf) << 8;
								W1 |= BBPR163;
							}
							

							WMatched = 0;
							// check valid Radar
							if (pDFSValidRadar->WLow)
							{
								if ( (W > (pDFSValidRadar->WLow - pDFSValidRadar->WMargin)) && 
								     (W < (pDFSValidRadar->WHigh + pDFSValidRadar->WMargin)) )
								{
									WMatched = 1;
								}
								
							}
							else
							{
								i = 0;
								while( (i < MAX_VALID_RADAR_W) && (pDFSValidRadar->W[i] != 0) )
								{
									if ( (W > (pDFSValidRadar->W[i] - pDFSValidRadar->WMargin)) &&
									     (W < (pDFSValidRadar->W[i] + pDFSValidRadar->WMargin)) )
									{
										WMatched = 1;
										break;
									}
									
									i++;
								}
							}
					
							//printk("WMatched == %d   ", WMatched);
							TMatched = 0;
							if (pDFSValidRadar->TLow)
							{
								if ( (T > (pDFSValidRadar->TLow - pDFSValidRadar->TMargin)) && 
								     (T < (pDFSValidRadar->THigh + pDFSValidRadar->TMargin)) )
								{
									TMatched = 1;
								}
							}
							else
							{
								i = 0;
								while( (i < MAX_VALID_RADAR_T) && (pDFSValidRadar->T[i] != 0) )
								{
									if ( (T > (pDFSValidRadar->T[i] - pDFSValidRadar->TMargin)) &&
									     (T < (pDFSValidRadar->T[i] + pDFSValidRadar->TMargin)) )
									{
										TMatched = 1;
										break;
									}
									
									i++;
								}
							}
							//printk("TMatched == %d\n", TMatched);

#ifdef DFS_DEBUG
							if (pAd->CommonCfg.DfsDebug.start == 2)
							{
								if ((T > (pAd->CommonCfg.DfsDebug.T_expected - pAd->CommonCfg.DfsDebug.T_margin)) && (T < (pAd->CommonCfg.DfsDebug.T_expected + pAd->CommonCfg.DfsDebug.T_margin)))
								{
									if (pAd->CommonCfg.DfsDebug.idx < 1000)
									{
										TestResult[pAd->CommonCfg.DfsDebug.idx].hit_time ++;
									}
								}
								else
								{
									if (pAd->CommonCfg.DfsDebug.idx < 1000)
									{
										TestResult[pAd->CommonCfg.DfsDebug.idx].false_time ++;
									}
								}
							}

#endif

							
						if ((TMatched == 1) && (WMatched == 1))
						{
							int k;

							UCHAR BBPR127;
							// found that seem all false detection are between 3000 ~ 10000
							if ((pAd->CommonCfg.ch_busy == 0) || ((T > 11000) && (pAd->CommonCfg.RadarDetect.RDDurRegion != CE)))
							{
								printk("%d W = %d, T= %d (%d)\n", jiffies, W, T, j);
								printk("radar detected!!\n");
								radarDeclared = 1;
							}
#ifdef DFS_DEBUG
							else
								printk("FDF: ch %d, w %d, t %d jf %d\n", j, W, T, jiffies);
							
							DBGPRINT(RT_DEBUG_TRACE, ("%d:%d,%d (%d)\n", pAd->CommonCfg.idle_time, pAd->CommonCfg.busy_time, pAd->CommonCfg.rssi, pAd->CommonCfg.ch_busy));
#endif
							if (pAd->CommonCfg.RadarReEnable == 0)
							{
								BBP_IO_WRITE8_BY_REG_ID(pAd, 140, BBPR140);
								printk("Disable detecting : write 140 = 0x%x\n", BBPR140);
								NewRadarDetectionStop(pAd);
								
							}
						}
						else if (TMatched == 1)
						{
							if ((pAd->CommonCfg.ch_busy == 0) || ((T > 11000) && (pAd->CommonCfg.RadarDetect.RDDurRegion != CE)))
							{
								printk("%d W = %d, T= %d (%d)\n", jiffies, W, T, j);
								if (pAd->CommonCfg.NewDfsDebug & RADAR_DEBUG_IGNORE_WIDTH)
								{
									printk("radar detected!!\n");
									radarDeclared = 1;
								}
								
							}
#ifdef DFS_DEBUG
							else
								printk("FDF: ch %d, w %d, t %d jf %d\n", j, W, T, jiffies);
							
							DBGPRINT(RT_DEBUG_TRACE, ("%d:%d,%d (%d)\n", pAd->CommonCfg.idle_time, pAd->CommonCfg.busy_time, pAd->CommonCfg.rssi, pAd->CommonCfg.ch_busy));
#endif
						}				
								pDFSValidRadar++;						
							
						}
							
					} // if (channel & (0x1 << j))
						
				} // for (j = 0;j < NEW_DFS_MAX_CHANNEL; j++)
			}

		} // if (channel)

		if (pAd->CommonCfg.RadarReEnable == 1)
		{
			BBP_IO_WRITE8_BY_REG_ID(pAd, 141, BBPR141);
		}

#ifdef DFS_DEBUG
		if (pAd->CommonCfg.McuRadarDebug & RADAR_SIMULATE)
		{
			radarDeclared = 1;
			pAd->CommonCfg.McuRadarDebug &= ~RADAR_SIMULATE;
		}
#endif
		

		if (radarDeclared)
		{
			// Radar found!!!
			if (!(pAd->CommonCfg.NewDfsDebug & RADAR_DONT_SWITCH))
			{
			
				for (i=0; i<pAd->ChannelListNum; i++)
				{
					if (pAd->CommonCfg.Channel == pAd->ChannelList[i].Channel)
					{
						pAd->ChannelList[i].RemainingTimeForUse = 1800;//30 min = 1800 sec
						break;
					}
				}

				pAd->CommonCfg.RadarDetect.ChMovingTime = 65;
			
				if (pAd->CommonCfg.RadarDetect.RDDurRegion == JAP_W56)
				{
					for (i = 0; i < pAd->ChannelListNum ; i++)
					{
						pAd->CommonCfg.Channel = APAutoSelectChannel(pAd, FALSE);
						if ((pAd->CommonCfg.Channel >= 100) && (pAd->CommonCfg.Channel <= 140))
							break;
					}
				}
				else if (pAd->CommonCfg.RadarDetect.RDDurRegion == JAP_W53)
				{
					for (i = 0; i < pAd->ChannelListNum ; i++)
					{
						pAd->CommonCfg.Channel = APAutoSelectChannel(pAd, FALSE);
						if ((pAd->CommonCfg.Channel >= 36) && (pAd->CommonCfg.Channel <= 60))
							break;
					}
				}
				else
					pAd->CommonCfg.Channel = APAutoSelectChannel(pAd, FALSE);
				
				N_ChannelCheck(pAd);
				if (pAd->CommonCfg.RadarDetect.RDMode =! RD_SILENCE_MODE)
				{
					pAd->CommonCfg.RadarDetect.RDMode = RD_SWITCHING_MODE;
					pAd->CommonCfg.RadarDetect.CSCount = 0;
				}
				else
				{
					APStop(pAd);
					APStartUp(pAd);
				}
				
				
			}
			
			// clear long pulse table
			pAd->CommonCfg.FCC_5[pAd->CommonCfg.fcc_5_idx].counter = 0;
			pAd->CommonCfg.fcc_5_idx = 0;
			pAd->CommonCfg.fcc_5_last_idx = 0;
			

		
		}
	}
	
	
}


void TimerCB(PRTMP_ADAPTER pAd);

VOID NewRadarDetectionStart(
	IN PRTMP_ADAPTER pAd)
{
	pNewDFSTable pDFS2Table;
	int i;
	UCHAR BBP_R140;
	UCHAR BBP_R142, BBP_R143, BBP_R144, BBP_R145, BBP_R146;
	UCHAR BBP_R147, BBP_R148, BBP_R149, BBP_R150, BBP_R151;
	UCHAR BBP_R152, BBP_R153, BBP_R154, BBP_R155;
		
	UCHAR BBPR4;

	BBP_IO_READ8_BY_REG_ID(pAd, BBP_R4, &BBPR4);

	if (pAd->CommonCfg.HtCapability.HtCapInfo.ChannelWidth  == BW_40)
	{
		BBP_IO_WRITE8_BY_REG_ID(pAd, 92, 3);
	}
	else
	{
		BBP_IO_WRITE8_BY_REG_ID(pAd, 91, 9);
	}
	
		
	DBGPRINT(RT_DEBUG_TRACE, ("NewRadarDetectionStart\n"));
	// choose Table
	if (pAd->CommonCfg.RadarDetect.RDDurRegion == FCC)
	{
		// Set up False detection filter
		pAd->CommonCfg.fdf_num = 3;
		// Rssi > -30, bandwith usage > 25%
		pAd->CommonCfg.ch_busy_threshold[0] = 2;
		pAd->CommonCfg.rssi_threshold[0] = -31;
		// Rssi > -35, bandwith usage > 50%
		pAd->CommonCfg.ch_busy_threshold[1] = 15;
		pAd->CommonCfg.rssi_threshold[1] = -36;
		// Rssi > -40, bandwith usage > 90%
		pAd->CommonCfg.ch_busy_threshold[2] = 90;
		pAd->CommonCfg.rssi_threshold[2] = -42;

		pDFS2Table = &NewDFSTable1[0];
	}
	else if (pAd->CommonCfg.RadarDetect.RDDurRegion == CE)
	{
		// Set up False detection filter
		pAd->CommonCfg.fdf_num = 3;
		// Rssi > -30, bandwith usage > 25%
		pAd->CommonCfg.ch_busy_threshold[0] = 2;
		pAd->CommonCfg.rssi_threshold[0] = -31;
		// Rssi > -35, bandwith usage > 50%
		pAd->CommonCfg.ch_busy_threshold[1] = 40;
		pAd->CommonCfg.rssi_threshold[1] = -36;
		// Rssi > -40, bandwith usage > 90%
		pAd->CommonCfg.ch_busy_threshold[2] = 90;
		pAd->CommonCfg.rssi_threshold[2] = -42;
		pDFS2Table = &NewDFSTable1[1];
	}
	else // Japan
	{
		// Set up False detection filter
		pAd->CommonCfg.fdf_num = 3;
		// Rssi > -30, bandwith usage > 25%
		pAd->CommonCfg.ch_busy_threshold[0] = 2;
		pAd->CommonCfg.rssi_threshold[0] = -31;
		// Rssi > -35, bandwith usage > 50%
		pAd->CommonCfg.ch_busy_threshold[1] = 15;
		pAd->CommonCfg.rssi_threshold[1] = -36;
		// Rssi > -40, bandwith usage > 90%
		pAd->CommonCfg.ch_busy_threshold[2] = 90;
		pAd->CommonCfg.rssi_threshold[2] = -42;

		pDFS2Table = &NewDFSTable1[2];
	}
	
	pAd->CommonCfg.MCURadarRegion = pDFS2Table->type;
	DBGPRINT(RT_DEBUG_TRACE, ("........pAd->CommonCfg.MCURadarRegion = %d\n", pAd->CommonCfg.MCURadarRegion));

	// delta delay
	BBP_IO_WRITE8_BY_REG_ID(pAd, 141, pAd->CommonCfg.DeltaDelay << 4);
	DBGPRINT(RT_DEBUG_TRACE, ("........Write 141 = 0x%x\n", pAd->CommonCfg.DeltaDelay << 4));
	
	if (pAd->CommonCfg.HtCapability.HtCapInfo.ChannelWidth  == BW_40)
		BBP_R140 = 0x80;
	else
		BBP_R140 = 0;
	
	
	// program channel i
	for (i = 0; i < NEW_DFS_MAX_CHANNEL; i++)
	{
		if (!((1 << i) & pAd->CommonCfg.ChEnable))
		{
			DBGPRINT(RT_DEBUG_TRACE, ("skip channel %d\n", i));
			continue;
		}
		BBP_R142 = 0;
		BBP_R142 |= (pDFS2Table->entry[i].avgLen & 0x100) >> 8;
		BBP_R142 |= (pDFS2Table->entry[i].TLow & 0x10000) >> 13;
		BBP_R142 |= (pDFS2Table->entry[i].THigh & 0x10000) >> 14;
		BBP_R142 |= (pDFS2Table->entry[i].mode & 0x7) << 4;
		
		BBP_R143 = pDFS2Table->entry[i].avgLen & 0xff;

		BBP_R144 = 0;
		BBP_R144 |= (pDFS2Table->entry[i].ELow & 0xf00) >> 4;
		BBP_R144 |= (pDFS2Table->entry[i].EHigh & 0xf00) >> 8;
		
		BBP_R145 = pDFS2Table->entry[i].ELow & 0xff;
		BBP_R146 = pDFS2Table->entry[i].EHigh & 0xff;

		BBP_R147 = (pDFS2Table->entry[i].TLow & 0xff00) >> 8;
		BBP_R148 = pDFS2Table->entry[i].TLow & 0xff;

		BBP_R149 = (pDFS2Table->entry[i].THigh & 0xff00) >> 8;
		BBP_R150 = pDFS2Table->entry[i].THigh & 0xff;

		BBP_R151 = pDFS2Table->entry[i].EpsilonT & 0xff;
		
		BBP_R152 = 0;
		BBP_R152 |= (pDFS2Table->entry[i].WLow & 0xf00) >> 4;
		BBP_R152 |= (pDFS2Table->entry[i].WHigh & 0xf00) >> 8;
		
		BBP_R153 = pDFS2Table->entry[i].WLow & 0xff;
		BBP_R154 = pDFS2Table->entry[i].WHigh & 0xff;
		
		BBP_R155 = pDFS2Table->entry[i].EpsilonW & 0xff;
		
		// select channel
		BBP_IO_WRITE8_BY_REG_ID(pAd, 140, BBP_R140 | (i << 4));
		DBGPRINT(RT_DEBUG_TRACE, ("........select channel, Write 140 = 0x%x\n", BBP_R140 | (i << 4)));
		
		// start programing
		BBP_IO_WRITE8_BY_REG_ID(pAd, 142, BBP_R142);
		BBP_IO_WRITE8_BY_REG_ID(pAd, 143, BBP_R143);
		BBP_IO_WRITE8_BY_REG_ID(pAd, 144, BBP_R144);
		BBP_IO_WRITE8_BY_REG_ID(pAd, 145, BBP_R145);
		BBP_IO_WRITE8_BY_REG_ID(pAd, 146, BBP_R146);
		BBP_IO_WRITE8_BY_REG_ID(pAd, 147, BBP_R147);
		BBP_IO_WRITE8_BY_REG_ID(pAd, 148, BBP_R148);
		BBP_IO_WRITE8_BY_REG_ID(pAd, 149, BBP_R149);
		BBP_IO_WRITE8_BY_REG_ID(pAd, 150, BBP_R150);
		BBP_IO_WRITE8_BY_REG_ID(pAd, 151, BBP_R151);
		BBP_IO_WRITE8_BY_REG_ID(pAd, 152, BBP_R152);
		BBP_IO_WRITE8_BY_REG_ID(pAd, 153, BBP_R153);
		BBP_IO_WRITE8_BY_REG_ID(pAd, 154, BBP_R154);
		BBP_IO_WRITE8_BY_REG_ID(pAd, 155, BBP_R155);
		DBGPRINT(RT_DEBUG_TRACE, ("........142 = 0x%x ", BBP_R142));
		DBGPRINT(RT_DEBUG_TRACE, ("143 = %x ", BBP_R143));
		DBGPRINT(RT_DEBUG_TRACE, ("144 = %x ", BBP_R144));
		DBGPRINT(RT_DEBUG_TRACE, ("145 = %x ", BBP_R145));
		DBGPRINT(RT_DEBUG_TRACE, ("146 = %x ", BBP_R146));
		DBGPRINT(RT_DEBUG_TRACE, ("147 = %x ", BBP_R147));
		DBGPRINT(RT_DEBUG_TRACE, ("148 = %x ", BBP_R148));
		DBGPRINT(RT_DEBUG_TRACE, ("149 = %x\n........", BBP_R149));
		DBGPRINT(RT_DEBUG_TRACE, ("150 = %x ", BBP_R150));
		DBGPRINT(RT_DEBUG_TRACE, ("151 = %x ", BBP_R151));
		DBGPRINT(RT_DEBUG_TRACE, ("152 = %x ", BBP_R152));
		DBGPRINT(RT_DEBUG_TRACE, ("153 = %x ", BBP_R153));
		DBGPRINT(RT_DEBUG_TRACE, ("154 = %x ", BBP_R154));
		DBGPRINT(RT_DEBUG_TRACE, ("155 = %x\n", BBP_R155));
	}
	
	// enable debug mode
	BBP_IO_WRITE8_BY_REG_ID(pAd, 126, 3);
	
	// Enable Radar detection
	BBP_IO_WRITE8_BY_REG_ID(pAd, 140, BBP_R140 | pAd->CommonCfg.ChEnable);
	DBGPRINT(RT_DEBUG_TRACE, ("........Write BBP 140 = 0x%x\n", BBP_R140 | pAd->CommonCfg.ChEnable));

	// Enable Bandwidth usage monitor
	{
		ULONG Value;
		RTMP_IO_READ32(pAd, CH_TIME_CFG, &Value);
		RTMP_IO_WRITE32(pAd, CH_TIME_CFG, Value | 0x1f);
	}


	request_tmr_service(NEW_DFS_WATCH_DOG_TIME, &TimerCB, pAd);
	
}


VOID NewRadarDetectionStop(
	IN PRTMP_ADAPTER pAd)
{
	DBGPRINT(RT_DEBUG_TRACE, ("NewRadarDetectionStop\n"));
	unregister_tmr_service();
	BBP_IO_WRITE8_BY_REG_ID(pAd, 140, 0x0);
	
	if (pAd->CommonCfg.HtCapability.HtCapInfo.ChannelWidth  == BW_40)
	{
		BBP_IO_WRITE8_BY_REG_ID(pAd, 92, 0);
	}
	else
	{
		BBP_IO_WRITE8_BY_REG_ID(pAd, 91, 4);
	}
}

#endif // NEW_DFS //

#endif // #if defined(DFS_SUPPORT) || defined(CARRIER_DETECTION_SUPPORT)

