/*
 * @file Member ID Definition of Group 3G
 *
 * Moderator: StanleyChu
 * Group ID: 0x001e0000/0x801e0000
 */

#ifndef _CSID_3G_H
#define _CSID_3G_H

#include "csid_gid.h"

#define _CFG_3G(x)		(CSID_GIDC_3G|(x))
#define _STA_3G(x)		(CSID_GIDS_3G|(x))

/* Extra Definitions */

#define CNT_TYPE_3G_ONDEMAND				0x00
#define CNT_TYPE_3G_AUTORECONNECT			0x01
#define CNT_TYPE_3G_MANUAL					0x02

#define CNT_STATUS_3G_DISCONNECT			0x00
#define CNT_STATUS_3G_CONNECTING			0x01
#define CNT_STATUS_3G_CONNECT				0x02
#define CNT_STATUS_3G_DISCONNECTING			0x03
#define CNT_STATUS_3G_WAIT_TRAFFIC			0x05
#define CNT_STATUS_3G_AUTH_FAIL				0x06
#define CNT_STATUS_3G_BACKUP				0x07
#define CNT_STATUS_3G_NOT_INSCHEDULE		0x08

#define HOTPLUG_STATUS_3G_HOLDING     		0x00
#define HOTPLUG_STATUS_3G_ADD     			0x01
#define HOTPLUG_STATUS_3G_REMOVE        	0x02
#define HOTPLUG_STATUS_3G_NONE				0x03
#define HOTPLUG_STATUS_3G_NOT_READY			0x04

#define ONDEMAND_STATUS_DISCONNECT_NO_IP	0x00
#define ONDEMAND_STATUS_CONNECT_NO_IP		0x01
#define ONDEMAND_STATUS_CONNECT_GET_IP		0x02

#define AUTHTYPE_3G_AUTO					0x00
#define AUTHTYPE_3G_PAP						0x01
#define AUTHTYPE_3G_CHAP					0x02

#define KEEP_ALIVE_3G_DISABLE				0x00
#define KEEP_ALIVE_3G_PING					0x01
#define KEEP_ALIVE_3G_LCP					0x02
#define KEEP_ALIVE_3G_PS_STATE				0x03
#define KEEP_ALIVE_3G_PING_MONITOR			0x04

#define ERR_CODE_3G_OK						0x00
#define ERR_CODE_3G_PIN_ERR					0x01
#define ERR_CODE_3G_PW_ERR					0x02
#define ERR_CODE_3G_NO_MODEM				0x03
#define ERR_CODE_3G_SETTING_ERR				0x05
#define ERR_CODE_3G_MODEM_NOT_READY			0x06
#define ERR_CODE_3G_NOT_INSCHEDULE			0x07
#define ERR_CODE_3G_SIM_NOT_INSERTED		0x08
#define ERR_CODE_3G_SIM_BLOCKED				0x09
#define ERR_CODE_3G_PIN_TRY_ERROR			0x0a
#define ERR_CODE_3G_AUTO_APN_ERROR			0x0b //no get imsi
#define ERR_CODE_3G_NO_SIGNAL    			0x0c //no signal

#define AUTOBAK_STATUS_3G_DISCONNECT		0x00
#define AUTOBAK_STATUS_3G_CONNECTING		0x01
#define AUTOBAK_STATUS_3G_CONNECT			0x02
#define AUTOBAK_STATUS_3G_DISCONNECTING		0x03

#define REG_STATUS_IDLE						0x00
#define REG_STATUS_REGISTERED				0x01
#define REG_STATUS_SEARCH					0x02
#define REG_STATUS_ROAMING					0x05

#define PROFILE_MLW							0x02
#define PROFILE_MOBILX						0x04
#define PROFILE_LIGHTSPEED					0x05
#define PROFILE_HARIER						0x06
#define PROFILE_DLINK_TELECO				0x07
#define PROFILE_TELEXCELL					0x08
#define PROFILE_LT_SVC						0x09
#define PROFILE_INDIA_HARIER				0x0a
#define PROFILE_TELEWELL_FINLAND			0x0b
#define PROFILE_TELEWELL_GERMANY			0x0c
#define PROFILE_COMPAT						0x0d
#define PROFILE_RAPIDMAX					0x0e
#define PROFILE_DLINK_CHINA					0x0f
#define PROFILE_IDENTECH					0x10
#define PROFILE_CMOTECH						0x11
#define PROFILE_NETCOMM						0x12
#define PROFILE_WEL							0x13
#define PROFILE_DLINK_DWR					0x14
#define PROFILE_DATACELL					0x15
#define PROFILE_ZONBEST						0x16
#define PROFILE_TECHNOSERV					0x17
#define PROFILE_ZOOM						0x18
#define PROFILE_TRENDNET			0x19

#define CONNECT_TYPE_PPPD					0x00
#define CONNECT_TYPE_DIRECTIP				0x01
#define CONNECT_TYPE_HSO					0x02
#define CONNECT_TYPE_IPHETH					0x03
#define CONNECT_TYPE_RNDIS					0x04
/* Configuration Items */
#define CSID_C_3G_USER          	_CFG_3G(0x0001)	//T=str, user name
#define CSID_C_3G_PASSWORD      	_CFG_3G(0x0002)	//T=str, user password
#define CSID_C_3G_APN	        	_CFG_3G(0x0003)	//T=str, APN
#define CSID_C_3G_PIN           	_CFG_3G(0x0004)	//T=str, PIN
#define CSID_C_3G_NUMBER        	_CFG_3G(0x0005)	//T=str, Dialed Number
#define CSID_C_3G_IF_IP         	_CFG_3G(0x0006)	//T=ipv4, static interface ip
#define CSID_C_3G_SERVICENAME   	_CFG_3G(0x0007)	//T=str, service name
#define CSID_C_3G_IF_PRIDNS     	_CFG_3G(0x0008)	//T=ipv4, dynamic interface primary DNS
#define CSID_C_3G_IF_SECDNS     	_CFG_3G(0x0009)	//T=ipv4, dynamic interface secondary DNS
#define CSID_C_3G_CNT_TYPE      	_CFG_3G(0x000a)	//T=u16, D=1, connect type, Dial-on-Demand, autoconnect...
#define CSID_C_3G_MTU           	_CFG_3G(0x000b)	//T=u16, D=0, max transmission unit
#define CSID_C_3G_IDLETIME      	_CFG_3G(0x000c)	//T=u16, D=600, max idle time
#define CSID_C_3G_AUTOBAK_ENABLE  	_CFG_3G(0x000d) //T=u16, 3g auto backup enable, 0:disable, 1:enable
#define CSID_C_3G_PING_HOST_IP		_CFG_3G(0x000e)	//T=ipv4, ping host IP for 3g auto backup
#define CSID_C_3G_AUTHTYPE			_CFG_3G(0x000f)	//T=u16, auth type, Auto, PAP, CHAP...
#define CSID_C_3G_KEEP_ALIVE		_CFG_3G(0x0010) //T=u16, 3g keep alive, disable, ping, lcp
#define CSID_C_3G_KA_PING_INTERVAL	_CFG_3G(0x0011)	//T=u16, D=60, ping interval for 3g keep alive
#define CSID_C_3G_KA_PING_IP		_CFG_3G(0x0012)	//T=ipv4, ping IP for 3g keep alive
#define CSID_C_3G_KA_LCP_INTERVAL	_CFG_3G(0x0013)	//T=u16, D=10, lcp interval for 3g keep alive 
#define CSID_C_3G_KA_LCP_TIMES		_CFG_3G(0x0014)	//T=u16, D=3, lcp times for 3g keep alive
#define CSID_C_3G_PREFER_MODE		_CFG_3G(0x0015)	//T=u16, D=0, 0:auto mode, 1:2g only, 2:3g only, 3:2g first, 4:3g first
#define CSID_C_3G_AUTO_APN          _CFG_3G(0x0016) //T=u16, D=1, 0:auto mode 1:Manual
#define CSID_C_3G_NAT_DISABLE		_CFG_3G(0x0017)	//T=u32, 3g wantype NAT disable
#define CSID_C_3G_USSD_RETURN		_CFG_3G(0x0018)	//T=str, ussd return str
#define CSID_C_3G_USSD				_CFG_3G(0x0019)	//T=str, ussd str
#define CSID_C_3G_PREFER_ENABLE		_CFG_3G(0x001a)	//T=u16, D=0, 3G prefer mode, 0:disable, 1:enable
#define CSID_C_3G_ROAMING_ENABLE	_CFG_3G(0x001B)	//T=u32, D=0, roaming switch, 0:disable, 1:enable
#define CSID_C_3G_PDP_TYPE          _CFG_3G(0x001c) //T=u16, D=0, 0:IP, 1=PPP
#define CSID_C_3G_CID               _CFG_3G(0x001D) //T=str, CID, D=1
#define CSID_C_3G_KA_PING_SIZE		_CFG_3G(0x001e)	//T=u16, D=56, ping packet size for 3g keep alive
#define CSID_C_3G_PING_REGULARLY_ENABLE		_CFG_3G(0x001f)	//T=u8, D=0, send ping regularly or not
#define CSID_C_3G_BDG_ETH_PORT		_CFG_3G(0x0020) //T=u16, 3g wan type bridge 2 ethernet port enable, 0:disable, 1:enable
#define CSID_C_3G_PING_INTERVAL     _CFG_3G(0x0021)	//T=u16, D=60, ping interval for 3g
#define CSID_C_3G_PING_IP		    _CFG_3G(0x0022)	//T=ipv4, ping IP for 3g
#define CSID_C_3G_PING_SIZE         _CFG_3G(0x0023)	//T=u16, D=56, ping packet size for 3g
#define CSID_C_3G_KEEP_ALIVE_FAIL_REBOOT    _CFG_3G(0x0024)	//T=u8, D=0, while keep alive fail: reboot or not
#define CSID_C_3G_KA_ERR_TIME       _CFG_3G(0x0025) //T=u16, D=4
#define CSID_C_3G_NOT_USE_DIRECTIP	_CFG_3G(0x0026)	//T=u16, D=0; 0:use direct IP; 1:not use direct IP
#define CSID_C_3G_PING_HOST			_CFG_3G(0x0027)	//T=str, ping domain or IP for 3g auto backup
#define CSID_C_3G_KA_PING_HOST		_CFG_3G(0x0028)	//T=str, ping domain or IP for 3g keep alive
#define CSID_C_3G_ROAMING_LED_ENABLE _CFG_3G(0x0029) //T=u32, D=0, roaming led switch, 0:disable, 1:enable
#define CSID_C_CHECK_RSSI_ENABLE 	_CFG_3G(0x002a) //T=u32, D=0, check rssi status 0:disable, 1:enable
#define CSID_C_3G_PROFILE			_CFG_3G(0x0030) //T=u16, for customer
#define CSID_C_3G_PRE_COMMAND		_CFG_3G(0x0031) //T=str, pre_command ,before connecting fill at command
#define CSID_C_3G_CNT_TYPE_UI      	_CFG_3G(0x0032)	//T=u16, D=1, connect type, Dial-on-Demand, autoconnect...
#define CSID_C_3G_MODULE_PWR_GPIO   _CFG_3G(0x0033)	//T=u8,  D=255, Power GPIO for 3G Module
#define CSID_C_3G_USE_RULE_NO		_CFG_3G(0x0040)	//T=u16, D=0,use_rule number
#define CSID_C_3G_SUPPORT_CARD_LIST	_CFG_3G(0x0050)	//T=str, 3g modem support list
#define CSID_C_3G_PIN_DEFAULT        _CFG_3G(0x0060) //T=str, PIN
#define CSID_C_3G_PIN_DEFAULT_RETRY  _CFG_3G(0x0061) //T=u32, D=0,
#define CSID_C_3G_SIERRA_NET_URB_SZ		 _CFG_3G(0x0062) //T=u32, D=0,
#define CSID_C_3G_PING_HOST_MODE		_CFG_3G(0x0063)	//T=u32, D=0,

/* Ping Monitor*/
#define CSID_C_3G_PING_MONITOR_PRIADDR       _CFG_3G(0x0070) //T=str, Ping Monitor primary address
#define CSID_C_3G_PING_MONITOR_SECADDR       _CFG_3G(0x0071) //T=str, Ping Monitor secondary address
#define CSID_C_3G_PING_MONITOR_TIMER         _CFG_3G(0x0072) //T=u32, Ping Monitor Timer
#define CSID_C_3G_PING_MONITOR_ACC_TIMER     _CFG_3G(0x0073) //T=u32, Ping Monitor accelerated Timer
#define CSID_C_3G_PING_MONITOR_ERR_COUNT     _CFG_3G(0x0074) //T=u32, Ping Monitor Fail Count

/*autobak/failover active(seamless) mode*/
#define CSID_C_3G_FAILOVER_MODE 	_CFG_3G(0x0080) //T=u32, 0: standby-mode 1:active(seamless)-mode

//for DNS Check for 3G connect
#define CSID_C_3G_DNS_CHECK_ENABLE          _CFG_3G(0x0090) //T=u8, D=0, 0:disbale 1:enable Switch to enable DNS check
#define CSID_C_3G_DNS_CHECK_INTERVAL        _CFG_3G(0x0091) //T=u32, D=30, Interval to DNS check
#define CSID_C_3G_DNS_CHECK_TARGET          _CFG_3G(0x0092) //T=str, target to query DNS
#define CSID_C_3G_DNS_CHECK_IDLE_MODE       _CFG_3G(0x0093) //T=u8, D=0, 0:disable 1:enable switch for enable dns check when in idle_

//ppp option
#define CSID_C_3G_PPP_OPTION		_CFG_3G(0x0100) //T=u32, ppp_option
#define CSID_C_3G_BAUDRATE			_CFG_3G(0x0101) //T=u32, D=0
#define CSID_C_3G_PEERDNS			_CFG_3G(0x0102) //T=str, D=0
#define CSID_C_3G_PPP_AUTHTYPE		_CFG_3G(0x0103)	//T=str, auth type, Auto, PAP, CHAP...
#define CSID_C_3G_PPP_LCP_INTERVAL	_CFG_3G(0x0104)	//T=str, lcp interval setting
#define CSID_C_3G_PPP_LCP_FAILURE	_CFG_3G(0x0105)	//T=str, lcp failure setting
#define CSID_C_3G_PPP_DEMAND		_CFG_3G(0x0106)	//T=str, demand option
#define CSID_C_3G_PPP_USERNAME		_CFG_3G(0x0107)	//T=str, username

#define CSID_C_3G_REMEBER_PIN		_CFG_3G(0x0200) //T=u16, 0:disable 1:enable
// 3g backup 3g 
#define CSID_C_3G_USE_PORT         	_CFG_3G(0x0800) //T=u16, 1:port1 2:port2
#define CSID_C_3G_S3G_ENABLE        _CFG_3G(0x0810) //T=u16, 0:disable 1:enable
#define CSID_C_3G_S3G_FAILOVER_MODE _CFG_3G(0x0820) //T=u32, 0: standby-mode 1:active-mode
//

#define CSID_C_3G_IMEI_NUMBER	_CFG_3G(0x0900) //T=str, IMEI for MEP Code, 8's word

#define CSID_C_3G_OP_LOCK_ENABLE	_CFG_3G(0x1000) //T=u16, for Operator Lock enable, 0:disable, 1:enable
#define CSID_C_3G_MCC0				_CFG_3G(0x1001) //T=str, for Mobile Contry Code
#define CSID_C_3G_MCC1				_CFG_3G(0x1002) //T=str, for Mobile Contry Code
#define CSID_C_3G_MCC2				_CFG_3G(0x1003) //T=str, for Mobile Contry Code
#define CSID_C_3G_MCC3				_CFG_3G(0x1004) //T=str, for Mobile Contry Code
#define CSID_C_3G_MCC4				_CFG_3G(0x1005) //T=str, for Mobile Contry Code
#define CSID_C_3G_MCC5				_CFG_3G(0x1006) //T=str, for Mobile Contry Code
#define CSID_C_3G_MCC6				_CFG_3G(0x1007) //T=str, for Mobile Contry Code
#define CSID_C_3G_MCC7				_CFG_3G(0x1008) //T=str, for Mobile Contry Code
#define CSID_C_3G_MNC0				_CFG_3G(0x1101) //T=str, for Mobile Network Code
#define CSID_C_3G_MNC1				_CFG_3G(0x1102) //T=str, for Mobile Network Code
#define CSID_C_3G_MNC2				_CFG_3G(0x1103) //T=str, for Mobile Network Code
#define CSID_C_3G_MNC3				_CFG_3G(0x1104) //T=str, for Mobile Network Code
#define CSID_C_3G_MNC4				_CFG_3G(0x1105) //T=str, for Mobile Network Code
#define CSID_C_3G_MNC5				_CFG_3G(0x1106) //T=str, for Mobile Network Code
#define CSID_C_3G_MNC6				_CFG_3G(0x1107) //T=str, for Mobile Network Code
#define CSID_C_3G_MNC7				_CFG_3G(0x1108) //T=str, for Mobile Network Code

//--- Multi 3G config CSID
#define CSID_C_3G_MULTIWAN_USER				 	_CFG_3G(0x2010)	//T=str, user name
#define CSID_C_3G_MULTIWAN_PASSWORD				_CFG_3G(0x2020)	//T=str, user password
#define CSID_C_3G_MULTIWAN_APN			    	_CFG_3G(0x2030)	//T=str, APN
#define CSID_C_3G_MULTIWAN_PIN				 	_CFG_3G(0x2040)	//T=str, PIN
#define CSID_C_3G_MULTIWAN_NUMBER				_CFG_3G(0x2050)	//T=str, Dialed Number
#define CSID_C_3G_MULTIWAN_IF_IP				_CFG_3G(0x2060)	//T=ipv4, static interface ip
#define CSID_C_3G_MULTIWAN_SERVICENAME		  	_CFG_3G(0x2070)	//T=str, service name
#define CSID_C_3G_MULTIWAN_IF_PRIDNS			_CFG_3G(0x2080)	//T=ipv4, dynamic interface primary DNS
#define CSID_C_3G_MULTIWAN_IF_SECDNS			_CFG_3G(0x2090)	//T=ipv4, dynamic interface secondary DNS
#define CSID_C_3G_MULTIWAN_CNT_TYPE				_CFG_3G(0x20a0)	//T=u16, D=1, connect type, Dial-on-Demand, autoconnect...
#define CSID_C_3G_MULTIWAN_MTU					_CFG_3G(0x20b0)	//T=u16, D=0, max transmission unit
#define CSID_C_3G_MULTIWAN_IDLETIME				_CFG_3G(0x20c0)	//T=u16, D=600, max idle time
#define CSID_C_3G_MULTIWAN_AUTOBAK_ENABLE		_CFG_3G(0x20d0) //T=u16, 3g auto backup enable, 0:disable, 1:enable
#define CSID_C_3G_MULTIWAN_PING_HOST_IP			_CFG_3G(0x20e0)	//T=ipv4, ping host IP for 3g auto backup
#define CSID_C_3G_MULTIWAN_AUTHTYPE				_CFG_3G(0x20f0)	//T=u16, auth type, Auto, PAP, CHAP...
#define CSID_C_3G_MULTIWAN_KEEP_ALIVE			_CFG_3G(0x2100) //T=u16, 3g keep alive, disable, ping, lcp
#define CSID_C_3G_MULTIWAN_KA_PING_INTERVAL		_CFG_3G(0x2110)	//T=u16, D=60, ping interval for 3g keep alive
#define CSID_C_3G_MULTIWAN_KA_PING_IP			_CFG_3G(0x2120)	//T=ipv4, ping IP for 3g keep alive
#define CSID_C_3G_MULTIWAN_KA_LCP_INTERVAL		_CFG_3G(0x2130)	//T=u16, D=10, lcp interval for 3g keep alive 
#define CSID_C_3G_MULTIWAN_KA_LCP_TIMES			_CFG_3G(0x2140)	//T=u16, D=3, lcp times for 3g keep alive
#define CSID_C_3G_MULTIWAN_PREFER_MODE			_CFG_3G(0x2150)	//T=u16, D=0, 0:auto mode, 1:2g only, 2:3g only, 3:2g first, 4:3g first
#define CSID_C_3G_MULTIWAN_AUTO_APN				_CFG_3G(0x2160) //T=u16, D=1, 0:auto mode 1:Manual
#define CSID_C_3G_MULTIWAN_NAT_DISABLE			_CFG_3G(0x2170)	//T=u32, 3g wantype NAT disable
#define CSID_C_3G_MULTIWAN_USSD_RETURN			_CFG_3G(0x2180)	//T=str, ussd return str
#define CSID_C_3G_MULTIWAN_USSD					_CFG_3G(0x2190)	//T=str, ussd str
#define CSID_C_3G_MULTIWAN_PREFER_ENABLE		_CFG_3G(0x21a0)	//T=u16, D=0, 3G prefer mode, 0:disable, 1:enable
#define CSID_C_3G_MULTIWAN_ROAMING_ENABLE		_CFG_3G(0x21b0)	//T=u32, D=0, roaming switch, 0:disable, 1:enable
#define CSID_C_3G_MULTIWAN_USE_PORT				_CFG_3G(0x21c0) //T=u32, 3g connect port
#define CSID_C_3G_MULTIWAN_BDG_ETH_PORT			_CFG_3G(0x2200) //T=u16, 3g wan type bridge 2 ethernet port enable, 0:disable, 1:enable
#define CSID_C_3G_MULTIWAN_PROFILE				_CFG_3G(0x2210) //T=u16, for customer
#define CSID_C_3G_MULTIWAN_USE_RULE_NO			_CFG_3G(0x2220)	//T=u16, D=0,use_rule number
#define CSID_C_3G_MULTIWAN_SUPPORT_CARD_LIST	_CFG_3G(0x2230)	//T=str, 3g modem support list

#define CSID_C_3G_MULTIWAN_OP_LOCK_ENABLE		_CFG_3G(0x2300) //T=u16, for Operator Lock enable, 0:disable, 1:enable
#define CSID_C_3G_MULTIWAN_MCC0					_CFG_3G(0x2310) //T=str, for Mobile Contry Code
#define CSID_C_3G_MULTIWAN_MCC1					_CFG_3G(0x2320) //T=str, for Mobile Contry Code
#define CSID_C_3G_MULTIWAN_MCC2					_CFG_3G(0x2330) //T=str, for Mobile Contry Code
#define CSID_C_3G_MULTIWAN_MCC3					_CFG_3G(0x2340) //T=str, for Mobile Contry Code
#define CSID_C_3G_MULTIWAN_MCC4					_CFG_3G(0x2350) //T=str, for Mobile Contry Code
#define CSID_C_3G_MULTIWAN_MCC5					_CFG_3G(0x2360) //T=str, for Mobile Contry Code
#define CSID_C_3G_MULTIWAN_MCC6					_CFG_3G(0x2370) //T=str, for Mobile Contry Code
#define CSID_C_3G_MULTIWAN_MCC7					_CFG_3G(0x2380) //T=str, for Mobile Contry Code
#define CSID_C_3G_MULTIWAN_MNC0					_CFG_3G(0x2410) //T=str, for Mobile Network Code
#define CSID_C_3G_MULTIWAN_MNC1					_CFG_3G(0x2420) //T=str, for Mobile Network Code
#define CSID_C_3G_MULTIWAN_MNC2					_CFG_3G(0x2430) //T=str, for Mobile Network Code
#define CSID_C_3G_MULTIWAN_MNC3					_CFG_3G(0x2440) //T=str, for Mobile Network Code
#define CSID_C_3G_MULTIWAN_MNC4					_CFG_3G(0x2450) //T=str, for Mobile Network Code
#define CSID_C_3G_MULTIWAN_MNC5					_CFG_3G(0x2460) //T=str, for Mobile Network Code
#define CSID_C_3G_MULTIWAN_MNC6					_CFG_3G(0x2470) //T=str, for Mobile Network Code
#define CSID_C_3G_MULTIWAN_MNC7					_CFG_3G(0x2480) //T=str, for Mobile Network Code

#define CSID_C_3G_PCTL_SWITCH					_CFG_3G(0x2490) //T=u32, pctl switch, 0:disable/1:enable
#define CSID_C_3G_MULTIWAN_PCTL_SWITCH     		_CFG_3G(0x2491) //T=u32, pctl switch, 0:disable/1:enable
#define CSID_C_3G_ATCOM_LOG_SWITCH     			_CFG_3G(0x2492) //T=u32, atcom log switch, 0:disable/1:enable
#define CSID_C_3G_SMS_SWITCH					_CFG_3G(0x24a0) //T=u32, sms switch , 0:disable/1:enable
#define CSID_C_3G_MULTIWAN_SMS_SWITCH     		_CFG_3G(0x24a1) //T=u32, sms switch , 0:disable/1:enable
#define CSID_C_3G_MULTIWAN_CNT_TYPE_UI          _CFG_3G(0x24a2) //T=u16, D=1, connect type, Dial-on-Demand, autoconnect...

#define CSID_C_3G_SAVE_PIN						_CFG_3G(0x24b0) //T=u32, save pin when pass , 0:disable/1:enable
#define CSID_C_3G_IMSI							_CFG_3G(0x24c0) //T=str, save imsi value
#define CSID_C_3G_RDIDX1						_CFG_3G(0x24d0) //T=u32, save read idx from 0-29
#define CSID_C_3G_RDIDX2						_CFG_3G(0x24e0) //T=u32, save read idx from 30-59
#define CSID_C_3G_RDIDX3						_CFG_3G(0x24e1) //T=u32, save read idx from 60-89
#define CSID_C_3G_RDIDX4						_CFG_3G(0x24e2) //T=u32, save read idx from 90-119
#define CSID_C_3G_PROXY_ENABLE					_CFG_3G(0x24f0) //T=u32, D=0, Proxy Enable, 0:disable, 1:enable
#define CSID_C_3G_PROXY_HOSTNAME				_CFG_3G(0x24f1) //T=str, Proxy Address
#define CSID_C_3G_PROXY_PORT					_CFG_3G(0x2521) //T=u16, Proxy Port (1-65535)
#define CSID_C_3G_BRIDGE_ENABLE					_CFG_3G(0x2531) //T=u32, D=0 ,0: 3G bridge disable ,1: 3G bridge enable
#define CSID_C_3G_CHECK_PREPAID_ENABLE          _CFG_3G(0x2532) //T=u32, D=0 ,0: Check SIM card prepaid value disable ,1: Check SIM card prepaid value enable
#define CSID_C_3G_SCHEDULE_REBOOT_DONE          _CFG_3G(0x2533) //T=u32, D=0 , 3G schedule reboot done 

#define CSID_C_3G_SMS_REMOTE_MG_ENABLE          _CFG_3G(0x2540) //T=u8, D=0 , SMS remote management enable, 0: disable, 1: enable
#define CSID_C_3G_DELETE_REMOTE_MG_SMS_ENABLE          _CFG_3G(0x2541) //T=u8, D=0 , Delete remote management SMS enable, 0: disable, 1: enable
#define CSID_C_3G_SMS_REMOTE_MG_SECURITY_KEY          _CFG_3G(0x2542) //T=str, D=0 , SMS remote management security key
#define CSID_C_3G_SMS_REMOTE_MG_CMD_STATUS_ENABLE          _CFG_3G(0x2543) //T=u8, D=0 , SMS remote management command status enable, 0: disable, 1: enable
#define CSID_C_3G_SMS_REMOTE_MG_CMD_CONNECT_ENABLE          _CFG_3G(0x2544) //T=u8, D=0 , SMS remote management command connect enable, 0: disable, 1: enable
#define CSID_C_3G_SMS_REMOTE_MG_CMD_DISCONNECT_ENABLE          _CFG_3G(0x2545) //T=u8, D=0 , SMS remote management command disconnect enable, 0: disable, 1: enable
#define CSID_C_3G_SMS_REMOTE_MG_CMD_RECONNECT_ENABLE          _CFG_3G(0x2546) //T=u8, D=0 , SMS remote management command reconnect enable, 0: disable, 1: enable
#define CSID_C_3G_SMS_REMOTE_MG_CMD_REBOOT_ENABLE          _CFG_3G(0x2547) //T=u8, D=0 , SMS remote management command reboot enable, 0: disable, 1: enable
#define CSID_C_3G_SMS_REMOTE_MG_NOTIFY_WAN_UP_ENABLE          _CFG_3G(0x2548) //T=u8, D=0 , SMS remote management notification wan up enable, 0: disable, 1: enable
#define CSID_C_3G_SMS_REMOTE_MG_NOTIFY_WAN_DOWN_ENABLE          _CFG_3G(0x2549) //T=u8, D=0 , SMS remote management notification wan down enable, 0: disable, 1: enable
#define CSID_C_3G_SMS_REMOTE_MG_NOTIFY_SEC_WAN_UP_ENABLE          _CFG_3G(0x254a) //T=u8, D=0 , SMS remote management notification secondary wan up enable, 0: disable, 1: enable
#define CSID_C_3G_SMS_REMOTE_MG_NOTIFY_SEC_WAN_DOWN_ENABLE          _CFG_3G(0x254b) //T=u8, D=0 , SMS remote management notification secondary wan down enable, 0: disable, 1: enable
#define CSID_C_3G_SMS_REMOTE_MG_ACCESS_CONTROL_ENABLE          _CFG_3G(0x254c) //T=u8, D=0 , SMS remote management access control enable, 0: disable, 1: enable
#define CSID_C_3G_SMS_REMOTE_MG_ACCESS_CONTROL_LIST          _CFG_3G(0x2550) //T=str, SMS remote management access control phone list
#define CSID_C_3G_SMS_REMOTE_MG_ACCESS_CONTROL_MG_ENABLE          _CFG_3G(0x2560) //T=u8, D=0, SMS remote management access control management enable, 0: disable, 1: enable
#define CSID_C_3G_SMS_REMOTE_MG_ACCESS_CONTROL_NOTIFY_ENABLE          _CFG_3G(0x2570) //T=u8, D=0, SMS remote management access control notification enable, 0: disable, 1: enable
#define CSID_C_3G_SMS_REMOTE_MG_CMD_PROCESS_IDX1          _CFG_3G(0x2580) //T=u32, D=0 , SMS remote management command process index from 0 ~ 29
#define CSID_C_3G_SMS_REMOTE_MG_CMD_PROCESS_IDX2          _CFG_3G(0x2581) //T=u32, D=0 , SMS remote management command process index from 30 ~ 59
#define CSID_C_3G_SMS_REMOTE_MG_CMD_PROCESS_IDX3          _CFG_3G(0x2582) //T=u32, D=0 , SMS remote management command process index from 60 ~ 89
#define CSID_C_3G_SMS_REMOTE_MG_CMD_PROCESS_IDX4          _CFG_3G(0x2583) //T=u32, D=0 , SMS remote management command process index from 90 ~ 119
#define CSID_C_3G_SMS_REMOTE_MG_ANY_PHONE_ENABLE          _CFG_3G(0x2590) //T=u8, D=0, SMS remote management any phone enable, 0: disable, 1: enable

#define CSID_C_3G_QMI_CURRENT_CLIENT_ID          _CFG_3G(0x2600) //T=u16, D=0, Current QMI connection client id

/* Status Items */
#define CSID_S_3G_CNT_TIME					_STA_3G(0x0100)	//T=u32, connect time
#define CSID_S_3G_MULTIWAN_CNT_TIME        	_STA_3G(0x0101)	//T=u32, connect time
#define CSID_S_3G_IF_IP						_STA_3G(0x0200)	//T=ipv4, dynamic interface ip
#define CSID_S_3G_MULTIWAN_IF_IP           	_STA_3G(0x0201)	//T=ipv4, dynamic interface ip
#define CSID_S_3G_IF_NM						_STA_3G(0x0300)	//T=ipv4, dynamic interface netmask
#define CSID_S_3G_MULTIWAN_IF_NM           	_STA_3G(0x0301)	//T=ipv4, dynamic interface netmask
#define CSID_S_3G_IF_GW						_STA_3G(0x0400)	//T=ipv4, dynamic interface gateway
#define CSID_S_3G_MULTIWAN_IF_GW           	_STA_3G(0x0401)	//T=ipv4, dynamic interface gateway
#define CSID_S_3G_IF_PRIDNS					_STA_3G(0x0500)	//T=ipv4, dynamic interface primary DNS
#define CSID_S_3G_MULTIWAN_IF_PRIDNS       	_STA_3G(0x0501)	//T=ipv4, dynamic interface primary DNS
#define CSID_S_3G_IF_SECDNS					_STA_3G(0x0600)	//T=ipv4, dynamic interface secondary DNS
#define CSID_S_3G_MULTIWAN_IF_SECDNS       	_STA_3G(0x0601)	//T=ipv4, dynamic interface secondary DNS
#define CSID_S_3G_CNT_STATUS				_STA_3G(0x0700)	//T=u16, connect status
#define CSID_S_3G_MULTIWAN_CNT_STATUS      	_STA_3G(0x0701)	//T=u16, connect status
#define CSID_S_3G_PKTIN_NUM					_STA_3G(0x0800)	//T=u32, number of packet input
#define CSID_S_3G_MULTIWAN_PKTIN_NUM       	_STA_3G(0x0801)	//T=u32, number of packet input
#define CSID_S_3G_PKTOUT_NUM				_STA_3G(0x0900)	//T=u32, number of packet output
#define CSID_S_3G_MULTIWAN_PKTOUT_NUM      	_STA_3G(0x0901)	//T=u32, number of packet output
#define CSID_S_3G_HOTPLUG_STATUS			_STA_3G(0x0a00) //T=u16, 3g hotplug status for script so it don't care for UI
#define CSID_S_3G_MULTIWAN_HOTPLUG_STATUS  	_STA_3G(0x0a01) //T=u16, 3g hotplug status for script so it don't care for UI
#define CSID_S_3G_ONDEMAND_STATUS			_STA_3G(0x0b00) //T=u16, demend wan type status for script so it don't care for UI
#define CSID_S_3G_MULTIWAN_ONDEMAND_STATUS 	_STA_3G(0x0b01) //T=u16, demend wan type status for script so it don't care for UI
#define CSID_S_3G_ALTERED					_STA_3G(0x0c00)	//T=u8, ui altered
#define CSID_S_3G_MULTIWAN_ALTERED         	_STA_3G(0x0c01)	//T=u8, ui altered
#define CSID_S_3G_CARD_INFO					_STA_3G(0x0d00)	//T=str, 3G card info
#define CSID_S_3G_MULTIWAN_CARD_INFO		_STA_3G(0x0d01)	//T=str, 3G card info
#define CSID_S_3G_SIGNAL_STRENGTH			_STA_3G(0x0e00)	//T=u16, signal strength
#define CSID_S_3G_MULTIWAN_SIGNAL_STRENGTH	_STA_3G(0x0e01)	//T=u16, signal strength
#define CSID_S_3G_SIGNAL_QUALITY			_STA_3G(0x0f00)	//T=str, signal quality
#define CSID_S_3G_MULTIWAN_SIGNAL_QUALITY	_STA_3G(0x0f01)	//T=str, signal quality
#define CSID_S_3G_NETWORK_NAME				_STA_3G(0x1000)	//T=str, network name
#define CSID_S_3G_MULTIWAN_NETWORK_NAME		_STA_3G(0x1001)	//T=str, network name
#define CSID_S_3G_ERR_CODE					_STA_3G(0x1100) //T=u16, connect fail error code
#define CSID_S_3G_MULTIWAN_ERR_CODE			_STA_3G(0x1101) //T=u16, connect fail error code
#define CSID_S_3G_AUTOBAK_STATUS			_STA_3G(0x1200)	//T=u16, 3g failover status
#define CSID_S_3G_MULTIWAN_AUTOBAK_STATUS	_STA_3G(0x1201)	//T=u16, 3g failover status
#define CSID_S_3G_SERVICE_TYPE				_STA_3G(0x1300)	//T=u16, 3g service type, 0:2G,1:none ,2:3G,3:4G
#define CSID_S_3G_MULTIWAN_SERVICE_TYPE		_STA_3G(0x1301)	//T=u16, 3g service type, 0:2G, 2:3G
#define CSID_S_3G_CARD_VENDORID				_STA_3G(0x1400)	//T=u16, 3G card vendor id
#define CSID_S_3G_MULTIWAN_CARD_VENDORID	_STA_3G(0x1401)	//T=u16, 3G card vendor id
#define CSID_S_3G_CARD_PRODUCTID			_STA_3G(0x1500)	//T=u16, 3G card product id
#define CSID_S_3G_MULTIWAN_CARD_PRODUCTID	_STA_3G(0x1501)	//T=u16, 3G card product id
#define CSID_S_3G_OP_LOCK_STATUS			_STA_3G(0x1600) //T=u16, for Operator Lock status, 0:unlock, 1:lock
#define CSID_S_3G_MULTIWAN_OP_LOCK_STATUS	_STA_3G(0x1601) //T=u16, for Operator Lock status, 0:unlock, 1:lock
#define CSID_S_3G_HOTPLUG_VERSION			_STA_3G(0x1700) //T=str, hotplug version
#define CSID_S_3G_MULTIWAN_HOTPLUG_VERSION	_STA_3G(0x1701) //T=str, hotplug version
#define CSID_S_3G_GET_STATUS				_STA_3G(0x1800) //T=u8, get_3g_satatus, 1:running
#define CSID_S_3G_MULTIWAN_GET_STATUS		_STA_3G(0x1801) //T=u8, get_3g_satatus, 1:running
#define CSID_S_3G_REG_STATUS				_STA_3G(0x1900) //T=u8, register status for 3G network
#define CSID_S_3G_MULTIWAN_REG_STATUS		_STA_3G(0x1901) //T=u8, register status for 3G network
#define CSID_S_3G_USSD_STATUS				_STA_3G(0x2300) //T=u8, D=0,get_ussd_satatus, 1:readable 0:unreadable
#define CSID_S_3G_MULTIWAN_USSD_STATUS		_STA_3G(0x2301) //T=u8, D=0,get_ussd_satatus, 1:readable 0:unreadable
#define CSID_S_3G_USB_PORT					_STA_3G(0x2b00) //T=u32, 1:PCIE 2:USB
#define CSID_S_3G_MULTIWAN_USB_PORT         _STA_3G(0x2b01) //T=u32, 1:PCIE 2:USB
#define CSID_S_3G_CMD_STATUS				_STA_3G(0x2c00) //T=u32, multiwan cmd status
#define CSID_S_3G_MULTIWAN_CMD_STATUS       _STA_3G(0x2c01) //T=u32, multiwan cmd status
#define CSID_S_3G_VOICE_REG_STATUS			_STA_3G(0x1a00) //T=u32, voice register status for 3G network

/* for 3G Failover ping check */
#define CSID_S_3G_WAN_IP					_STA_3G(0x2000) //T=ipv4, wan ip
#define CSID_S_3G_MULTIWAN_WAN_IP			_STA_3G(0x2001) //T=ipv4, wan ip
#define CSID_S_3G_WAN_GW					_STA_3G(0x2100) //T=ipv4, wan gateway
#define CSID_S_3G_MULTIWAN_WAN_GW			_STA_3G(0x2101) //T=ipv4, wan gateway
#define CSID_S_3G_WAN_IF_NAME				_STA_3G(0x2200) //T=str, the interface name of wan, ppp0, eth1, ixp1
#define CSID_S_3G_MULTIWAN_WAN_IF_NAME		_STA_3G(0x2201) //T=str, the interface name of wan, ppp0, eth1, ixp1
#define CSID_S_3G_IPSEC_ALTERED				_STA_3G(0x2b10) //T=u8, wan connect alter for vpn
#define CSID_S_3G_MULTIWAN_IPSEC_ALTERED    _STA_3G(0x2b11) //T=u8, wan connect alter for vpn

//---SMS function added by York 090916
#define CSID_S_3G_SMS_RECEIVED				_STA_3G(0x2400) //T=u8, D=0 ,numbers of reveived messages
#define CSID_S_3G_MULTIWAN_SMS_RECEIVED     _STA_3G(0x2401) //T=u8, D=0 ,numbers of reveived messages
#define CSID_S_3G_SMS_NEW					_STA_3G(0x2500) //T=u8, D=0 ,numbers of new messages
#define CSID_S_3G_MULTIWAN_SMS_NEW          _STA_3G(0x2501) //T=u8, D=0 ,numbers of new messages
#define CSID_S_3G_SMS_CAPACITY				_STA_3G(0x2600) //T=u8, D=0 ,numbers of SIM card messages capacity
#define CSID_S_3G_MULTIWAN_SMS_CAPACITY     _STA_3G(0x2601) //T=u8, D=0 ,numbers of SIM card messages capacity
#define CSID_S_3G_SMS_TOPHONE				_STA_3G(0x2700) //T=str, phone number of sending
#define CSID_S_3G_MULTIWAN_SMS_TOPHONE      _STA_3G(0x2701) //T=str, phone number of sending
#define CSID_S_3G_SMS_TEXT					_STA_3G(0x2800) //T=str, text content of sending
#define CSID_S_3G_MULTIWAN_SMS_TEXT         _STA_3G(0x2801) //T=str, text content of sending
#define CSID_S_3G_SMS_ERR_MSG				_STA_3G(0x2900) //T=u8,	send error messages 
#define CSID_S_3G_MULTIWAN_SMS_ERR_MSG      _STA_3G(0x2901) //T=u8,	send error messages 
#define CSID_S_3G_SMS_STATUS				_STA_3G(0x2a00) //T=u8, D=0 ,sms status , 0:querying, 1:query finished
#define CSID_S_3G_MULTIWAN_SMS_STATUS       _STA_3G(0x2a01) //T=u8, D=0 ,sms status , 0:querying, 1:query finished
#define CSID_S_3G_SMS_QUERY_FLAG			_STA_3G(0x2c00) //T=u8, sms query flay, let commander know 3G connect or not
#define CSID_S_3G_MULTIWAN_SMS_QUERY_FLAG   _STA_3G(0x2c01) //T=u8, sms query flay, let commander know 3G connect or not
#define CSID_S_3G_SMS_FULL_ALERT			_STA_3G(0x2c02) //T=u8, D=0 ,sms full alert. if sms full D=1
#define CSID_S_3G_SMS_RETURN_STATUS			_STA_3G(0x2c03) //T=u8, D=0 ,sms return status. if send ok D=1, if send fail D=2
#define CSID_S_3G_SMS_REMOTE_MG_PROCESS_CONNECTION			_STA_3G(0x2c04) //T=u8, D=0 ,sms remote management processing wan connection . 0: do nothing, 1: stop commander wan process

//---Pin Control Corresponding Parameters, added by York 091117
#define CSID_S_3G_PCTL_ACTION					_STA_3G(0x2d00) //T=u32, D=0 ,1:disable 2:enable 3:change 4:unlock_pin 5:unlock_puk
#define CSID_S_3G_MULTIWAN_PCTL_ACTION			_STA_3G(0x2d01) //T=u32, D=0 ,1:disable 2:enable 3:change 4:unlock_pin 5:unlock_puk
#define CSID_S_3G_PCTL_LOCK_PIN					_STA_3G(0x2e00) //T=str,SIM lock(pin code)
#define CSID_S_3G_MULTIWAN_PCTL_LOCK_PIN		_STA_3G(0x2e01) //T=str,SIM lock(pin code)
#define CSID_S_3G_PCTL_CHG_OLD_PIN				_STA_3G(0x2f00) //T=str,chagne function, old password
#define CSID_S_3G_MULTIWAN_PCTL_CHG_OLD_PIN		_STA_3G(0x2f01) //T=str,chagne function, old password
#define CSID_S_3G_PCTL_CHG_NEW_PIN				_STA_3G(0x3000) //T=str,change function, new password
#define CSID_S_3G_MULTIWAN_PCTL_CHG_NEW_PIN		_STA_3G(0x3001) //T=str,change function, new password
#define CSID_S_3G_PCTL_CHG_VRF_PIN				_STA_3G(0x3100) //T=str,change function, use to verify new password again
#define CSID_S_3G_MULTIWAN_PCTL_CHG_VRF_PIN		_STA_3G(0x3101) //T=str,change function, use to verify new password again
#define CSID_S_3G_PCTL_UNLOCK_PIN				_STA_3G(0x3200) //T=str,if SIM lock(pin), use to unlock
#define CSID_S_3G_MULTIWAN_PCTL_UNLOCK_PIN		_STA_3G(0x3201) //T=str,if SIM lock(pin), use to unlock
#define CSID_S_3G_PCTL_UNLOCK_PUK				_STA_3G(0x3300) //T=str,if SIM lock(puk), use to unlock
#define CSID_S_3G_MULTIWAN_PCTL_UNLOCK_PUK		_STA_3G(0x3301) //T=str,if SIM lock(puk), use to unlock
#define CSID_S_3G_PCTL_UNLOCK_NEW_PIN			_STA_3G(0x3400) //T=str,if SIM lock(puk), use to unlock
#define CSID_S_3G_MULTIWAN_PCTL_UNLOCK_NEW_PIN  _STA_3G(0x3401) //T=str,if SIM lock(puk), use to unlock
#define CSID_S_3G_PCTL_PIN_RETRY_TIMES			_STA_3G(0x3500) //T=s32, pin retry times, -1 means the module isn't supported
#define CSID_S_3G_MULTIWAN_PCTL_PIN_RETRY_TIMES _STA_3G(0x3501) //T=s32, pin retry times, -1 means the module isn't supported
#define CSID_S_3G_PCTL_PUK_RETRY_TIMES			_STA_3G(0x3600) //T=s32, puk retry times, -1 means the module isn't supported
#define CSID_S_3G_MULTIWAN_PCTL_PUK_RETRY_TIMES _STA_3G(0x3601) //T=s32, puk retry times, -1 means the module isn't supported
#define CSID_S_3G_PCTL_STATUS					_STA_3G(0x3700) //T=u32, 0:querying, 1:query finished
#define CSID_S_3G_MULTIWAN_PCTL_STATUS          _STA_3G(0x3701) //T=u32, 0:querying, 1:query finished
#define CSID_S_3G_PCTL_ERR_MSG					_STA_3G(0x3800) //T=s32, pin control error messages
#define CSID_S_3G_MULTIWAN_PCTL_ERR_MSG         _STA_3G(0x3801) //T=s32, pin control error messages
#define CSID_S_3G_PCTL_UI_SHOW					_STA_3G(0x3900) //T=s32, for UI for distinguish next step
#define CSID_S_3G_MULTIWAN_PCTL_UI_SHOW         _STA_3G(0x3901) //T=s32, for UI for distinguish next step
#define CSID_S_3G_PCTL_QUERY_FLAG				_STA_3G(0x3a00) //T=u32, pctl query flag, let commander know 3G connect or not
#define CSID_S_3G_MULTIWAN_PCTL_QUERY_FLAG      _STA_3G(0x3a01) //T=u32, pctl query flag, let commander know 3G connect or not
#define CSID_S_3G_PCTL_BTN						_STA_3G(0x3b00) //T=u32, pctl button flag, 1:disable/2:enable
#define CSID_S_3G_MULTIWAN_PCTL_BTN    			_STA_3G(0x3b01) //T=u32, pctl button flag, 1:disable/2:enable
#define CSID_S_3G_NETWORK_TYPE					_STA_3G(0x3e00) //T=str, network type , NONE/GPRS/EDGE/UMTS/HSDPA/HSUPA
#define CSID_S_3G_MULTIWAN_NETWORK_TYPE    		_STA_3G(0x3e01) //T=str, network type , NONE/GPRS/EDGE/UMTS/HSDPA/HSUPA
#define CSID_S_3G_AUTOBAK_MODE					_STA_3G(0x4000) //T=u16, 0:local wan connected, 1:local wan disconnected
#define CSID_S_3G_CONNECT_TYPE					_STA_3G(0x4001) //T=u16,D=0 0:ppp, 1:sierra_dhcp, 2:hso_dhcp, 3:ipheth, 4: rndis, 5:qmi_dhcp  
#define CSID_S_3G_SIM_CARD_STATUS				_STA_3G(0x5000) //T=str, sim card status
#define CSID_S_3G_SIM_CARD_STATUS_HIJACKING		_STA_3G(0x5001) //T=u32, sim card status for Hijacking

#define CSID_S_3G_MODEM_FSN						_STA_3G(0x5100) //T=str, modem Factory Serial Number
#define CSID_S_3G_MODEM_TEMPERATURE				_STA_3G(0x5200) //T=str, modem Temperature
#define CSID_S_3G_SYSTEM_MODE					_STA_3G(0x5201) //T=str, 3g system mode
#define CSID_S_3G_WCDMA_BAND					_STA_3G(0x5202) //T=str, 3g WCDMA band
#define CSID_S_3G_WCDMA_CHANNEL					_STA_3G(0x5203) //T=str, 3g WCDMA channel
#define CSID_S_3G_GMM_STATE						_STA_3G(0x5204) //T=str, 3g GMM (PS) state
#define CSID_S_3G_MM_STATE						_STA_3G(0x5205) //T=str, 3g MM (CS) state
#define CSID_S_3G_RX_SIGNAL_STRENGTH			_STA_3G(0x5206) //T=str, 3g RX Signal Strength
#define CSID_S_3G_GET_EXTRA						_STA_3G(0x5207) //T=u8, 3g GET EXTRA
#define CSID_S_3G_IMSI							_STA_3G(0x5300) //T=str, 3g IMSI
#define CSID_S_3G_MODEM_HW_REVISION				_STA_3G(0x5400) //T=str, modem hardware revision
#define CSID_S_3G_COUNTRY_CODE					_STA_3G(0x5500) //T=str, 3g country code
#define CSID_S_3G_NETWORK_CODE					_STA_3G(0x5600) //T=str, 3g network code
#define CSID_S_3G_CELL_ID						_STA_3G(0x5700) //T=str, Cell ID
#define CSID_S_3G_PRIMARY_SCRAMBLING_CODE		_STA_3G(0x5800) //T=str, Primary Scrambling Code
#define CSID_S_3G_HSUPA_CATEGORY				_STA_3G(0x5900) //T=str, HSUPA category
#define CSID_S_3G_HSDPA_CATEGORY				_STA_3G(0x5a00) //T=str, HSDPA category
#define CSID_S_3G_RSCP							_STA_3G(0x5b00) //T=str, Received Signal Code Power
#define CSID_S_3G_MEP_STATUS					_STA_3G(0x6000) //T=u32, MEP_status
#define CSID_S_3G_MEP_ERR_MSG					_STA_3G(0x6001) //T=u32, MEP error message
#define CSID_S_3G_MCC_MNC						_STA_3G(0x6002) //T=str, MCC MNC
#define CSID_S_3G_MEP_CLCK						_STA_3G(0x6003) //T=str, MEP CLCK Status
#define CSID_S_3G_MODEM_RESET_TIME				_STA_3G(0x6004) //T=u8, 3G modem RESER time
#define CSID_S_3G_CNT_SCHEDULE_OFF				_STA_3G(0x7000) //T=u32,
#define CSID_S_3G_SIM_PREPAID_STATUS    		_STA_3G(0x7010) //T=u32, D=0, 3G SIM card prepaid status
#define CSID_S_3G_AUTO_APN_STATUS				_STA_3G(0x7100) //T=u32, D=0, 1:success -1: fail
#endif //ifndef _CSID_3G_H
